#include "dfs.hh"

// fuse_ino_t and ino_t are both 8 bytes (on 64 bit system)

DFS *dfs;


DFS::DFS(string volume) {
  zero_frame = 0;
  verify_assumptions();
  base = NULL; // must be set by first event
  dfs = this;
  pthread_mutex_init(&pending_lock,NULL);
  pthread_mutex_init(&replay_lock,NULL);
  pthread_mutex_init(&fuse_lock,NULL);
  block_dev = volume;
  frame_counter = 0;
  inode_id_counter = 0;
  segment_sequence_counter = 0;
  log_sequence_counter = 1; // object 0 is invalid
  nframes = fs_get_size(block_dev) / SEGMENT_SIZE_BYTES;
  assert(nframes > 0);
}

void DFS::commit_pending_log(bool write_all_segments) {
  pthread_mutex_lock(&pending_lock);
  list<Segment *>::iterator segment_iterator;
  Segment *segment;

  list<Segment *> full_segments;

  int debug_counter = 0;
  while(pending_log_entries.size()) {
    Event *e = *pending_log_entries.begin();
    debug_counter++;
    pending_log_entries.pop_front();
    select_open_segment(e->get_size(), segment_iterator);
    segment = *segment_iterator;
    segment->append(e);
    if(segment->full()) {
      full_segments.push_front(segment);
      open_segments.erase(segment_iterator);
    }

  }
      
  // now we've added all pending header entries, 
  // it's time to add pending data objects
  
  // cout << "2. Handle pending data objects\n";
  //flush(cout);
  
  list<DataObject *>::iterator j;
  DataObject *o;
  segment_iterator = open_segments.begin();

  // FIXME write object summaries!
  while(pending_data_entries.size() > 0) {
    j = pending_data_entries.begin();
    o = *j;
    pending_data_entries.erase(j);
    j = pending_data_entries.begin();

    //cout << "   data object " << o << "\n";
    //flush(cout);
    
    if(segment_iterator == open_segments.end()) {
      if(open_segments.size() == 0) {
	segment_id new_segment = next_segment_id();
	open_segments.push_back(new Segment(new_segment));
	segment_logical_to_physical[new_segment] = nframes;
      }
      segment_iterator = open_segments.begin();
    }
    segment = *segment_iterator;
   

    if(o->get_size() > segment->free_space) {
      DataObject *n = o->split(segment->free_space, next_object_id());
      n->event_id = next_object_id();
      o->next = n->event_id;
      pending_data_entries.push_back(n);
      assert(o->get_size() == segment->free_space);
    }

    event_summary summary;
    segment->append(o, &summary);
    object_info[o->event_id] = summary;

    if(segment->full()) {
      segment_iterator = open_segments.erase(segment_iterator);
      full_segments.push_back(segment);
    }
  }
  
  time_t now = time(NULL);
  
  segment_iterator = open_segments.begin();
  while(segment_iterator != open_segments.end()) {
    segment = *segment_iterator;
    if(write_all_segments || segment->full() || (now - segment->creation > SEGMENT_FRESHNESS_TIMEOUT)) {
      full_segments.push_back(segment);
      segment_iterator = open_segments.erase(segment_iterator);
    } else
      segment_iterator++;
  }
  

  segment_iterator = full_segments.begin();

  while(segment_iterator != full_segments.end()) {
    segment = *segment_iterator;
    segment_iterator = full_segments.erase(segment_iterator);
    frame_id segment_frame = empty_segment_frame();
    //cout << "   write segment " << segment->header->id << " to, frame " << segment_frame << "\n"; flush(cout);
    assert(segment_frame != NoFreeSegments); // disk full; not implemented
    assert(segment->write(segment_frame) == 0);
    
    map<segment_id,frame_id>::iterator itr = segment_logical_to_physical.find(segment->header->id);

    assert(itr != segment_logical_to_physical.end());
    segment_logical_to_physical.erase(itr);
    segment_logical_to_physical[segment->header->id] = segment_frame;
    delete segment;
  }
  
  pthread_mutex_unlock(&pending_lock);    
}

int DFS::format_volume(string volume, Event *base) {
  return format_volume(volume, base, false);
}

void verify_assumptions() {
  if(sizeof(fuse_ino_t) != sizeof(ino_t)) {
    cerr << "WARNING sizeof(fuse_ino_t) != sizeof(ino_t)\n";
    cerr << "   What, is this a 32-bit machine?!\n";
  }
}

bool is_log_event_type(entry_type t) {
  return (t != LOG_ERROR && t != LOG_DATA);
}



void DFS::read_volume() {

  vector<event_summary> log;
  
  for(frame_id frame = 0; frame < nframes; frame++) {
    Segment s(frame, true);
    if(s.header->id > 0) {
      segment_sequence_counter = std::max(segment_sequence_counter,s.header->id);
      frame_counter = std::max(frame+1,frame_counter);
      segment_id seg = s.header->id;
      segment_logical_to_physical[seg] = frame;
      s.read_table();
      for(int i = 0; i < s.header->n_objects; i++) {
	segment_header_entry *obj = &s.table[i];
	
	event_summary summary = {obj->id, seg, obj->offset, obj->type, obj->size};
	log_sequence_counter = std::max(obj->id,log_sequence_counter);
	object_info[obj->id] = summary;
	
	if(is_log_event_type(obj->type)) {
	  log.push_back(summary);
	}
      }
    }
  }

  //  cout << "Done reading volume.  " << log.size() << " log events found\n";
  //cout << "log sequence counter = " << log_sequence_counter << "\n";
  //cout << "segment_sequence_counter = " << segment_sequence_counter << "\n";
  //cout << "Sorting...\n";

  sort(log.begin(), log.end(), event_summary_comparison());

  cout << "Replaying log\n";
  unsigned int start_event;
  for(start_event = log.size()-1; start_event > 0; start_event--) {
    if(log[start_event].type == LOG_ZERO) { 
      frame_counter = segment_sequence_counter - log[start_event].segment + 1;
      break;
    }
  }
  
  for(unsigned int i = start_event; i < log.size(); i++) {
    replay_log(log[i]);
  }
  
  
  // make sure the next object counter is working
  for(map<inode_id,Summary*>::iterator j = summaries.begin(); j != summaries.end(); j++)
    use_inode_id(j->first);
  
  cout << "----------------- done replaying log -------------------------\n";
}

template <class T> 
void DFS::replay_serialized(event_summary &summary) {

  T logobj;  
  logobj.event_id = summary.id;

  char *buffer = new char[summary.size];
  
  read_segment( (void *) buffer, summary.segment, summary.offset, summary.size); // open ifstream
  istringstream netstrm(string(buffer,summary.size), ios_base::binary | ios_base::in);
  BOOST_ARCHIVE_ITYPE ia(netstrm);
  ia >> logobj;
  delete[] buffer;
  logobj.replay();
}

void DFS::replay_log(event_summary &summary) { 

  //  cout << "Replay " << eventname(summary.type) << "   (event " << summary.id << ")\n";

  switch(summary.type) 
    { 
    case LOG_ERROR: 
    case LOG_DATA: 
      cout << "error: trying to replay invalid log event type " << summary.type << "\n";
      break;
    case LOG_TEST:
    case LOG_TEST_DATA:
      cout << "(replaying test events does nothing)\n";
      break;
    case LOG_MAP_INODE:
      replay_serialized<LogMapINode>(summary);
      break;
    case LOG_SETATTR:
      replay_serialized<LogSetattr>(summary);
      break;     
    case LOG_CREATE_REGULAR_FILE:
      replay_serialized<LogCreateRegularFile>(summary);
      break;     
    case LOG_SET_DIRENT:
      replay_serialized<LogSetDirent>(summary);
      break;
    case LOG_CREATE_DIRECTORY:
      replay_serialized<LogCreateDirectory>(summary);
      break;
    case LOG_WRITE:
      replay_serialized<LogWrite>(summary);
      break;
    case LOG_CACHE:
      replay_serialized<LogCache>(summary);
      break;
    case LOG_POSIX_BASE:
      replay_serialized<LogUsePOSIXBase>(summary);
      break;
    case LOG_S3RO_BASE:
      replay_serialized<LogUseS3ROBase>(summary);
      break;
    case LOG_S3RO_TYPE:
      replay_serialized<LogSetS3ROType>(summary);
      break;
    case LOG_ZERO:
      replay_serialized<LogZero>(summary);
      break;
    default:
      cout << "attempt to replay unrecognized log event type" << summary.type << "\n";
      assert(false);
    }
}

int DFS::format_volume(string volume, Event *base, bool wipe) {
  // if store is empty, initialize structures
  // otherwise

  off_t volume_size = fs_get_size(volume);
  cout << "Init " << volume << ", " << volume_size << " bytes\n";
  long long n = volume_size / (SEGMENT_SIZE_BYTES);

  cout << "Number of segments " << n << "\n";

  FILE *f = fopen(volume.c_str(),"r+");

  if(f == NULL) {
    cout << "error opening file " << errno << "\n";
    return errno;
  }

  cout << "Formatting volume\n";

  segment_header empty_header;
  empty_header.id = 0;
  empty_header.n_objects = 0;
  
  char *cbuf = new char[SEGMENT_SIZE_BYTES];
  void *buffer = (void *) cbuf;
  memset(buffer, 0, SEGMENT_SIZE_BYTES);
  
  for(int i = 0; i < n; i++) {

    if(PERCENT(i,n)%5 != PERCENT(i-1,n)%5) 
      cout << "  " << PERCENT(i,n) << "%  ";
    
    if(fseek(f, frame_offset(i), SEEK_SET)) {
      cerr << "error with fseek " <<  errno << "\n";
      return 0;
    }

    if(!wipe) {
      if( !fwrite(&empty_header, sizeof(segment_header), 1, f) ) {
	cerr << "error with fwrite " <<  errno << "\n";
	return 0;
      }
    } else {
      if(!fwrite(buffer, SEGMENT_SIZE_BYTES, 1, f)) {
	cerr << "error with fwrite " <<  errno << "\n";
	return 0;
      }
    }

  }
  
  fclose(f);
  cout << "\n";

  DFS dfs(volume);
  base->play();
  dfs.baptize(dfs.get_root_ino(), dfs.base->root_path);

  // get root inode from base filesystem
  //dfs.baptize(dfs.get_root_ino(), dfs.base->root_path);
  dfs.destroy();

  delete[] cbuf;
  return 0;
}

void Event::play() {
  dfs->append_log_object(this);
  this->replay();
  dfs->commit_pending_log(false);
}

Summary *DFS::get_summary(inode_id n) {
  return get_summary(n, true);
}

Summary *DFS::get_summary(inode_id n, bool assert_fail) {
  map<inode_id, Summary *>::iterator i = summaries.find(n);
  if(assert_fail)
    assert(i != summaries.end());
  else if(i == summaries.end())
    return NULL;
  Summary *s = i->second;
  if(s->type == STUB_SUMMARY) {
    Summary *summary = ((StubSummary *)s)->complete();
    summaries.erase(i);
    delete s;
    summaries[n] = summary;
    return summary;
  } else {
    return s;
  }
}

void DFS::delete_summary(inode_id n) {
  map<inode_id, Summary *>::iterator i = summaries.find(n);
  assert(i != summaries.end());
  Summary *s = i->second;
  summaries.erase(i);
  delete s;
}


/* man 2 path_resolution */
/*	 * Valid replies:
	 *   fuse_reply_entry
	 *   fuse_reply_err
	 */
void DFS::dfs_lookup(fuse_req_t req, fuse_ino_t parent, const char *name) {

  if(strcmp(name,"FREEZE") == 0) {
    cout << "Freeze command received...\n";
    fuse_reply_err(req, ENOENT);
    freeze();
    return;
  }

  struct fuse_entry_param e;
  memset(&e,0,sizeof(e));


  Summary *parent_summary = get_summary(parent);

  assert(parent_summary->type == DIRECTORY_SUMMARY);
  DirectorySummary *parent_directory = (DirectorySummary *) parent_summary;

  //  cout << "autofail: lookup " << name << " in " << parent_directory->base_path << "\n";
  //fuse_reply_err(req,ENOENT);
  //return;


  // search for name in the dirent diff map
  map<string,inode_id>::iterator found_name;
  Summary *summary = NULL; 

  errno = 0;
  inode_id child = parent_directory->lookup(name);
  if(child == DELETED || child == 0) {
    //    if(errno)
     // fuse_reply_err(req,errno);
    //else
    fuse_reply_err(req,ENOENT);
    return;
  }

  summary = get_summary(child);
  e.ino = child;
  summary->getattr(&e.attr);

  fuse_reply_entry(req, &e);
}

inode_id DirectorySummary::lookup(string name) {

  if(name == string(".")) return my_id;

  assert(name != string(".."));

  map<string,inode_id>::iterator found_name;
  found_name = dirent_diff.find(name);
  if(found_name != dirent_diff.end()){ 
    return found_name->second;
  }
  else {
    return base_lookup(name);
  }
}

inode_id DirectorySummary::base_lookup(string name) {
  //cout << "  [Directory " << base_path << " base lookup " << name << " ...]\n";
  if(not in_base) return 0;
  return dfs->base->lookup(base_path, name);
}

void Summary::getattr(struct stat *attr) {
  if(mstat != NULL) {
    memcpy(attr, mstat, sizeof(struct stat));
  } else {
    assert(in_base);
    assert(dfs->base->get_stat(base_path, attr)==0);
    attr->st_ino = my_id;
  }
}

void Summary::setattr(struct stat *attr, object_id creator) {
  if(mstat == NULL) {
    mstat = new struct stat;
    assert(mstat != NULL);
  }
  mstat_creator = creator;
  memcpy(mstat, attr, sizeof(struct stat));
}

void FileSummary::setattr(struct stat *attr, object_id creator) {
  struct stat s;
  getattr(&s);
  //if(mstat == NULL) {
  //  mstat = new struct stat;
  //  assert(mstat != NULL);
  //}
  mstat_creator = creator;
  memcpy(mstat, attr, sizeof(struct stat));
  
  if(s.st_size != attr->st_size) 
    change_size(s.st_size, attr->st_size, creator);
}


void FileSummary::change_size(size_t oldsize, size_t newsize, object_id creator) {
  if(newsize > oldsize) {
    // pad with zeroes
    //    cout << "ZERO PAD oldsize =" << oldsize << " newsize " << newsize  << "  pad size" <<  (newsize-oldsize) << "\n";
    Overlay *pad = Overlay::zeroes(oldsize, newsize-oldsize, creator);
    add_overlay(pad);
  }
  else {
    //    cout << "TRUNCATE oldsize " << oldsize << "  newsize " << newsize << "\n";
    // truncate
    Overlay temp(newsize,0,0,0,0);
    vector<Overlay *>::iterator i = lower_bound(content_diff.begin(), content_diff.end(), &temp, OverlayPointerComparison());
    // i is first overlay with offset >= newsize
    
    // delete overlays greater than new size
    while(i != content_diff.end()) {
      Overlay *x = *i;
      delete x;
      i = content_diff.erase(i);
    }
    if(i != content_diff.begin()) {
      i--;
      Overlay *prev = *i;
      assert(prev->offset < (off_t) newsize);
      if(prev->offset + prev->size > (off_t) newsize)
	prev->size = newsize- prev->offset;
    }
  }
}

/*------------------- fuse functions ------------------------ */

/* man 2 stat ( also check http://en.wikipedia.org/wiki/Stat_(Unix) ) */
void DFS::dfs_getattr(fuse_req_t req, fuse_ino_t ino, struct fuse_file_info *fi) {
  struct stat attr;
  Summary *s = get_summary(ino);
  s->getattr(&attr);
  fuse_reply_attr(req, &attr,0);
}


/* man 2 stat, man 2 chmod */
void DFS::dfs_setattr(fuse_req_t req, fuse_ino_t ino, struct stat *attr,
		 int to_set, struct fuse_file_info *fi) {
  struct stat newattr;
  Summary *s = get_summary(ino);
  flush(cout);
  s->getattr(&newattr);  

  if(to_set & FUSE_SET_ATTR_MODE) 
    newattr.st_mode = attr->st_mode;
  if(to_set & FUSE_SET_ATTR_UID) 
    newattr.st_uid = attr->st_uid;
  if(to_set & FUSE_SET_ATTR_GID) 
    newattr.st_gid = attr->st_gid;
  if(to_set & FUSE_SET_ATTR_SIZE) {
    // FIXME handle size change correctly
    // SERIOUS FIXME
    cout << "<< FUSE_SET_ATTR_SIZE >>\n";
    if(newattr.st_size != attr->st_size)
      cout << "   size has changed " << newattr.st_size << "  --> " << attr->st_size << "\n";
    else
      cout << "   no size change?\n";
    newattr.st_size = attr->st_size;
  }
  if(to_set & FUSE_SET_ATTR_ATIME) 
    newattr.st_atime = attr->st_atime;
  if(to_set & FUSE_SET_ATTR_MTIME) 
    newattr.st_mtime = attr->st_mtime;

  (new LogSetattr(ino, newattr))->play();
  fuse_reply_attr(req, &newattr,0);
}



/* man 2 mknod */
void DFS::dfs_mknod(fuse_req_t req, fuse_ino_t parent, const char *name,
		    mode_t mode, dev_t rdev) {
  if(!S_ISREG(mode)) {
    cerr << " only regular files and directories are supported\n";
    // EPERM error if not supported
    fuse_reply_err(req, EPERM);
    return;
  }
  cout << "[dfs_mknod " << name << "]\n";

  struct fuse_entry_param e;
  memset(&e,0,sizeof(e));

  // 1. create new file inode
  inode_id id = next_inode_id();

  time_t now = time(NULL);
  uid_t uid = geteuid();
  gid_t gid = getegid();

  // construct attributes  
  //(other fields not relevant for reg. files)
  struct stat *s = &e.attr;
  s->st_ino = id;
  s->st_mode = mode;
  s->st_nlink = 0;
  s->st_uid = uid;
  s->st_gid = gid;
  s->st_size = 0;
  s->st_atime = now;
  s->st_mtime = now;
  s->st_mtime = now;
  s->st_ctime = now;

  (new LogCreateRegularFile(id, e.attr))->play();

  e.ino = id;

  // 2. create parent directory entry
  (new LogSetDirent(string(name), parent, id))->play();

  fuse_reply_entry(req, &e);
}

/* man 2 mkdir */
void DFS::dfs_mkdir(fuse_req_t req, fuse_ino_t parent, const char *name,
		    mode_t mode) {

  struct fuse_entry_param e;
  memset(&e,0,sizeof(e));

  // 1. create new dir inode
  inode_id id = next_inode_id();

  time_t now = time(NULL);
  uid_t uid = geteuid();
  gid_t gid = getegid();

  // construct attributes  
  //(other fields not relevant for reg. files)
  struct stat *s = &e.attr;
  s->st_ino = id;
  s->st_mode = mode | S_IFDIR | 0755;
  s->st_nlink = 2;
  s->st_uid = uid;
  s->st_gid = gid;
  s->st_size = 4096; // arbitrary
  s->st_atime = now;
  s->st_mtime = now;
  s->st_mtime = now;
  s->st_ctime = now;

  (new LogCreateDirectory(id, e.attr))->play();

  e.ino = id;

  // 2. create parent directory entry
  (new LogSetDirent(string(name), parent, id))->play();

  fuse_reply_entry(req, &e);  
}

/* man 2 unlink */
void DFS::dfs_unlink(fuse_req_t req, fuse_ino_t parent, const char *name) {
  Summary *parent_summary = get_summary(parent);
  if(parent_summary->type != DIRECTORY_SUMMARY) {
    fuse_reply_err(req, ENOTDIR);
    return;
  }
  DirectorySummary *parent_directory = (DirectorySummary *) parent_summary;
  
  inode_id child = parent_directory->lookup(name);

  if(child == DELETED || child == 0) {
    fuse_reply_err(req, ENOENT);
    return;
  }
  
  // FIXME other error conditions need to be handled? ,  see man 2 unlink

  (new LogSetDirent(string(name), parent, DELETED))->play();
  fuse_reply_err(req, 0);
}

/* man 2 rmdir */
void DFS::dfs_rmdir(fuse_req_t req, fuse_ino_t parent, const char *name) {
  Summary *parent_summary = get_summary(parent);
  if(parent_summary->type != DIRECTORY_SUMMARY) {
    fuse_reply_err(req, ENOTDIR);
    return;
  }
  DirectorySummary *parent_directory = (DirectorySummary *) parent_summary;
  
  inode_id child = parent_directory->lookup(name);
  if(child == DELETED || child == 0) {
    fuse_reply_err(req, ENOENT);
    return;
  }

  Summary *csummary = get_summary(child);
  if(csummary->type != DIRECTORY_SUMMARY) {
    fuse_reply_err(req, ENOTDIR);
    return;
  }

  DirectorySummary *d = (DirectorySummary *) csummary;

  if(!d->empty()) {
    fuse_reply_err(req, ENOTEMPTY);
    return;
  }

  (new LogSetDirent(string(name), parent, DELETED))->play();
  fuse_reply_err(req, 0);
}


/* man 2 rename --- annoyingly intricate
 * in principle, like unlink/rmdir + mknod/mkdir modulo corner cases Fuse 
 * takes care of the fact that newparent may not be a subdirectory of parent.
 */
void DFS::dfs_rename(fuse_req_t req, fuse_ino_t parent, const char *name,
		     fuse_ino_t newparent, const char *newname) {
  // FIXME many errors not handled
  Summary *parent_summary = get_summary(parent);
  if(parent_summary->type != DIRECTORY_SUMMARY) {
    fuse_reply_err(req, ENOTDIR);
    return;
  }
  DirectorySummary *parent_directory = (DirectorySummary *) parent_summary;

  inode_id original = parent_directory->lookup(name);
  
  if(original == DELETED) {
    fuse_reply_err(req, ENOENT);
    return;
  }
  
  (new LogSetDirent(string(name), parent, DELETED))->play();
  (new LogSetDirent(string(newname), newparent, original))->play();
  
  fuse_reply_err(req,0);
}


/* man 2 read */
void DFS::dfs_read(fuse_req_t req, fuse_ino_t ino, size_t size, off_t off, struct fuse_file_info *fi) {

  if(size == 0) {
    fuse_reply_buf(req, NULL, 0);
    return;
  }

  Summary *s = get_summary(ino);
  if(s->type != FILE_SUMMARY) {
    fuse_reply_err(req, EISDIR);
    return;
  }
  FileSummary *f = (FileSummary *) s;
  
  char buffer[size];
  size_t actual_size = f->read(buffer, off, size);
  fuse_reply_buf(req, buffer, actual_size);
}

/* man 2 write, also check ``man 2 lseek'' for out of bounds write */
void DFS::dfs_write(fuse_req_t req, fuse_ino_t ino, const char *buf,
		    size_t size, off_t off, struct fuse_file_info *fi) {
  if(size == 0) {
    fuse_reply_write(req,0);
    return;
  }

  Summary *s = get_summary(ino);
  if(s->type != FILE_SUMMARY) {
    fuse_reply_err(req, EISDIR);
    return;
  }

  // FIXME no bounds check for overwriting size!
  // SUPER FIXME no error condition handling

  DataObject *data = new DataObject(string(buf,size));
  
  (new LogWrite(ino, off, size, append_data_object(data)))->play();
  
  fuse_reply_write(req,size);
}


/* man 2 readdir */
void DFS::dfs_readdir(fuse_req_t req, fuse_ino_t ino, size_t size, off_t off,
		      struct fuse_file_info *fi) {

  //  cout << "readdir " << ino << "\n";

  Summary *s = get_summary(ino);

  //  cout << "    base " << s->base_path << "\n";

  if(s->type != DIRECTORY_SUMMARY) {
    fuse_reply_err(req, ENOTDIR);
    return;
  }

  DirectorySummary *d = (DirectorySummary *) s;

  struct dirbuf b;
  memset(&b, 0, sizeof(b));
  //dirbuf_add(req, &b, ".", 1);
  //dirbuf_add(req, &b, "..", 1);

  //cout << "  compute directory stream\n";

  d->directory_stream(req,b);

  //cout << "  done\n";

  reply_buf_limited(req, b.p, b.size, off, size);
  free(b.p);
}

void DFS::dfs_destroy(void *userdata) {
  dfs->destroy();
}


void Segment::read_table() {
  assert(header_read);
  if(table_read) 
    return;
  table_read = true;
  if(header->n_objects == 0) 
    return;
  long table_bytes = sizeof(segment_header_entry) * header->n_objects;
  table = (segment_header_entry *) (PTR_ADD(header,-table_bytes));

  segment_offset header_offset = SEGMENT_SIZE_BYTES - sizeof(segment_header);
  segment_offset table_offset = header_offset - table_bytes;
  
  assert(frame < dfs->nframes); // this function should not be called on open_segments, which have frame = nframes

  dfs->read_frame(table, frame, table_offset, table_bytes);

  //  assert(fseek(f,table_start, SEEK_SET) == 0);
  //assert(fread(table, table_bytes, 1, f) == 1);
  
  if(header->n_objects > 0) {
    cout << "first object id " << table->id << " offset " << table->offset << " type " << table->type << "\n";
  }
}

Segment::Segment(frame_id fr, bool foo) {
  // foo is not used for anything, it allows us to distinguish
  // this constructor from the other

  // read from file
  data_read = false;
  frame = fr;
  table_read = false;
  header_read = true;
  table = NULL;

  segment_offset header_offset = SEGMENT_SIZE_BYTES - sizeof(segment_header);

  header = (segment_header *) &(data[header_offset]);

  dfs->read_frame(header, frame, header_offset, sizeof(segment_header));

  //  assert(fseek(f,SEGMENT_SIZE_BYTES-sizeof(segment_header),SEEK_CUR) == 0);
  //assert(fread(header, sizeof(segment_header), 1, f) == 1);

}
  
int Segment::write(frame_id fr) {
  dfs->write_frame(data, fr);
  return 0;
}

/*string Segment::fetch_object(unsigned int index) {
  read_table();
  assert(index < header->n_objects);
  unsigned char *ptr = &(data[table[index].offset]);
  dfs->read_frame(ptr, frame, table[index].offset, table[index].size);
  return string( (const char *) ptr,table[index].size);
  }*/

#define IT vector<Overlay *>::iterator




void FileSummary::add_overlay(Overlay *o) {
  Overlay *x;
  //cout << "-----------\n";
  //  cout << "INSERT offset " << o->offset << " size " << o->size << "   data " << o->data << "  data offset " << o->data_offset << "\n";
  //cout << "    *** before\n";
  //dump_content();

  assert(o->creator != 0);
	 
  IT i = lower_bound(content_diff.begin(), content_diff.end(), o, OverlayPointerComparison());

  i = content_diff.insert(i,o);

  // truncate previous overlay if necessary

  if(i != content_diff.begin()) {
    IT prev(i);
    prev--;
    x = *prev;
    assert(x->offset < o->offset); // based on understanding of how lower_bound works
    // previous overlay is non-empty
    if(x->extent() > o->offset) {
      if(x->extent() > o->extent()) {
	// new overlay is completely contained in an old overlay
	// split the old overlay in to two pieces, one for each side
	Overlay *x2 = new Overlay(x);
	x2->offset = o->extent();
	x2->data_offset += o->extent() - x->offset;
	x->truncate(o->offset);
	content_diff.insert(i,x2);
	add_overlay(o);
	return;
      }
      else {
	x->truncate(o->offset);
      }
    }
  }
  
  IT next(i);
  next++;

  while(next != content_diff.end()) {
    x = *next;
    if(x->offset < o->extent()) {
      if(x->extent() < o->extent()) {
	next = content_diff.erase(next); // totally eclipsed by overlay
	delete x;
      } else {
	x->ftruncate(o->extent());
	if(x->size == 0) {
	  next = content_diff.erase(next); 
	  delete x;
	}
      }
    } else {
      break;
    }
  }
};
#undef IT

void LogSetattr::replay() { 
  BEGIN_REPLAY;
  //  cout << "[replay LogSetattr for " << ino << "]\n";
  struct stat a;
  memcpy(&a, attr.c_str(), sizeof(struct stat));
  dfs->get_summary(ino)->setattr(&a, event_id);
  END_REPLAY;
}

void LogMapINode::replay() { 
  BEGIN_REPLAY;
  StubSummary *s = new StubSummary(id, base_path.c_str(), event_id);
  
  dfs->use_inode_id(id);

  // sanity check: assert id is not already in summaries
  map<ino_t,inode_id>::iterator j = dfs->inode_map.find(base_ino);
  assert(j == dfs->inode_map.end());

  map<inode_id,Summary*>::iterator i = dfs->summaries.find(id);
  assert(i == dfs->summaries.end());
  // --
  
  dfs->summaries[id] = s;
  dfs->inode_map[base_ino] = id;
  END_REPLAY;
}

inode_id DFS::baptize(ino_t base_inode, string base_path) {
  assert(base_path.size() > 0);

  assert(base_inode != 0);
  map<ino_t, inode_id>::iterator temp = inode_map.find(base_inode);
  if(temp != inode_map.end()) {
    // probably a hard link
    return temp->second;
  }

  inode_id new_id = next_inode_id();
  (new LogMapINode(base_inode, new_id, base_path))->play();

  return new_id;
}

Summary *StubSummary::complete() {
  struct stat attr;
  getattr(&attr);
  mode_t m = attr.st_mode;
  if(S_ISREG(m)) {
    return new FileSummary(my_id, base_path, creator);
  } else if(S_ISDIR(m)) {
    return new DirectorySummary(my_id, base_path, creator);
  } else if(S_ISLNK(m)) {
    // not implemented
    assert(false);
  } else if(S_ISCHR(m)) {
    // not implemented
    assert(false);
  } else if(S_ISBLK(m)) {
    // not implemented
    assert(false);
  } else if(S_ISFIFO(m)) {
    // not implemented
    assert(false);
  } else if(S_ISSOCK(m)) {   
    // not implemented
    assert(false);
  } else { 
    // ?????
    assert(false);
  }
}

void LogCreateRegularFile::replay() {
  BEGIN_REPLAY;
  struct stat attr;
  getattr(attr);
  // create summary
  // sanity check
  map<inode_id,Summary*>::iterator i = dfs->summaries.find(file_id);
  assert(i == dfs->summaries.end());
  Summary *s = new FileSummary(file_id, &attr, event_id);
  dfs->summaries[file_id] = s;
  END_REPLAY;
}

void LogCreateDirectory::replay() {
  BEGIN_REPLAY;
  struct stat attr;
  getattr(attr);
  dfs->summaries[directory_id] = new DirectorySummary(directory_id, &attr, event_id);
  END_REPLAY;
}


void LogSetDirent::replay() {
  BEGIN_REPLAY;

  Summary *s1 = dfs->get_summary(parent);
  assert(s1->type == DIRECTORY_SUMMARY);
  DirectorySummary *parent_summary = (DirectorySummary *) s1;
  // cout << "[Replay SetDirent " << parent << "(" << parent_summary->base_path << ") " << name << " = " << id << "]\n";
  if(id != DELETED) {

    Summary *child_summary  = dfs->get_summary(id, false);
    if(child_summary == NULL) { 
      cout << "    id " << id << " get_summary failed!!\n";
      return; // bummer
    }
    struct stat attr;

    child_summary->getattr(&attr);
    if(!attr.st_mode & S_IFDIR) { // necessary?
      attr.st_nlink++;
      child_summary->setattr(&attr,event_id); 
    }
  } 
  else { 
    // entry is being deleted 
    inode_id old = parent_summary->lookup(name.c_str());
    if(old != 0) {
      // there is a previous entry
      Summary *old_summary = dfs->get_summary(old);
      struct stat oldattr;
      old_summary->getattr(&oldattr);
      if(!oldattr.st_mode & S_IFDIR) {
	oldattr.st_nlink--;
	old_summary->setattr(&oldattr, event_id);
	if(oldattr.st_nlink <= 0) 
	  dfs->delete_summary(old);
      }
      //	cout << "  Note could save memory by deleting summary with 0 hard links!!\n";
    }
  }
  parent_summary->set_entry(name, id, event_id);
  END_REPLAY;
}

bool DirectorySummary::empty() {
  // annoyingly complicated
  map<string,inode_id>::iterator i;
  for(i = dirent_diff.begin(); i != dirent_diff.end(); i++)
    if(i->second != DELETED) return false;

  return base_empty();
}

bool DirectorySummary::base_empty() {
  if(not in_base) return true;

  list<string> entries;

  dfs->base->ls(base_path,entries);

  for(list<string>::iterator j = entries.begin(); j != entries.end(); j++) {
    map<string,inode_id>::iterator i = dirent_diff.find(*j);
    if(i != dirent_diff.end() && i->second != DELETED)
      return false;
  }

  return true;
}

#define diffmap vector<Overlay *>

size_t FileSummary::read(char *buffer, offset_t offset, size_t size) { 
  // buffer is exactly size bytes
  // returns bytes read (<= size)
  // request may go beyond file

  struct stat attr;
  getattr(&attr);

  cout << "[Read " << my_id << "  offset " << offset << " bytes " << size << "  file size " << attr.st_size << "]\n";

  if(offset >= attr.st_size)
    return 0;
  if( (off_t) (offset + size) > attr.st_size) 
    size = attr.st_size - offset;
  
  size_t bytes_read = 0;
  diffmap::iterator i = content_diff.begin();

  Overlay o(0,0,0,0,0);
  Overlay *x;
  Overlay *containing_overlay;
  offset_t next_boundary; // start offset of next boundary
  while(size > 0) {
    o.offset = offset;
    containing_overlay = NULL;
    next_boundary = attr.st_size;
    i = lower_bound(i, content_diff.end(), &o, OverlayPointerComparison());
    if(i != content_diff.end()) {
      x = *i;
      // overlay x is either after the data block we're looking for,
      // or the start of it
      
      if(x->offset == offset) {
	// x is the beginning of the block we want
	containing_overlay = x;
	next_boundary = x->extent();
      } else {
	// x is after the block we want; examine previous overlay
	next_boundary = x->offset;
	if(i != content_diff.begin()) {
	  // there exists a previous overlay
	  i--;
	  x = *i;
	  assert(x->offset < offset);
	  if(offset < x->offset + x->size) {
	    // offset is contained in x
	    containing_overlay = x;
	    next_boundary = x->extent();
	  } 
	}
      }
    } 
    size_t chunk_size = std::min(size, (size_t)(next_boundary-offset)); 
    cout << "   chunk size " << chunk_size << "  size " << size << " next bound " << next_boundary << "  offset " << offset << "\n";
    assert(chunk_size >= 0);

    size_t bread = 0; // loop bytes read
    
    if(chunk_size > 0) {
      if(containing_overlay != NULL) {
	// read chunk from overlay

	cout << "       source: overlay data=" << containing_overlay->data << "\n";
	bread =  DataObject(containing_overlay->data).read(buffer, 
							   offset-containing_overlay->offset,
							   chunk_size);
      } else {
	cout << "       source: base\n";
	bread = read_base_chunk(buffer, offset, chunk_size);


      }
      size -= bread;
      bytes_read += bread;
      offset += bread;

      /* debugging begin */
      string tempread;
      for(int tempi = 0; tempi < std::min( (int) bread,30); tempi++){
	if(buffer[tempi] >= 33 && buffer[tempi] <=  126) 
	  tempread += string(buffer[tempi],1);
	else
	  tempread += string("X");
      }
      cout << "       " << tempread << "\n";

      /*debugging end */
      

      buffer = (char *) PTR_ADD(buffer,bread);
    }
  }
  return bytes_read;
}

#undef diffmap

size_t DataObject::read(void *buffer, offset_t offset, size_t size) {
  if(my_id == 0) {
    // object zero is a magical zero-padding data object
    memset(buffer,0,size);
    return size;
  }

  dfs->commit_pending_log();
  load_summary();
  if(offset >= summary.size)
    // refer to next segment
    return DataObject(get_next()).read(buffer,offset - summary.size, size);
  
  size_t effective_size = std::min((long int) (summary.size-offset),(long int) size);

  dfs->read_segment(buffer, summary.segment, summary.offset + sizeof(object_id), effective_size);

  if(size > effective_size) {
    cout <<"      handoff to data " << get_next() << "\n";
    // pass on to next data object
    return effective_size + DataObject(get_next()).read(PTR_ADD(buffer, effective_size), 0, size-effective_size);
  }
  else
    return effective_size;
}

object_id DataObject::get_next() {
  dfs->commit_pending_log();
  load_summary();
  dfs->read_segment(&next, summary.segment, summary.offset, sizeof(object_id));
  return next;
}

frame_id DataObject::get_frame() {
  load_summary();
  map<segment_id,frame_id>::iterator i = dfs->segment_logical_to_physical.find(summary.segment);
  assert(i != dfs->segment_logical_to_physical.end());
  return i->second;
}

void DataObject::load_summary() {
  assert(my_id != 0);
  dfs->commit_pending_log(false);
  if(!summary_cached) {
    map<object_id, event_summary>::iterator i;
    i = dfs->object_info.find(my_id);
    assert(i != dfs->object_info.end());
    summary = i->second;
    assert(summary.type == LOG_DATA);
    summary_cached = true;
  }
}


void DirectorySummary::directory_stream(fuse_req_t req, struct dirbuf &b) {

  fill_dirent_diff();

  for(map<string,inode_id>::iterator i = dirent_diff.begin(); i != dirent_diff.end(); i++) {
    inode_id child = i->second;
    if(child == DELETED) continue;
    Summary *s = dfs->get_summary(child);
    struct stat attr;
    s->getattr(&attr);
    //    cout << "    dirbuf add " << i->first.c_str() << "\n";
    dirbuf_add(req, &b, i->first.c_str(), child);
  }
}


void DirectorySummary::fill_dirent_diff() {
  if(!in_base || dirent_diff_filled) 
    return;

  //  cout << "   fill dirent diff\n";
  
  dirent_diff_filled = true;

  list<string> names;
  list<ino_t> inos;
  list<string> paths;

  dfs->base->ls(base_path, names, inos, paths);

  map<string,inode_id>::iterator i;

  list<ino_t>::iterator ino_i = inos.begin();
  list<string>::iterator path_i = paths.begin();

  for(list<string>::iterator name_i = names.begin(); name_i != names.end(); name_i++, ino_i++, path_i++) {
    i = dirent_diff.find(*name_i);
    if(i == dirent_diff.end())
      dirent_diff[*name_i] = dfs->baptize(*ino_i, *path_i);
  }
}

void LogWrite::replay() {
  BEGIN_REPLAY;

  struct stat attr;
  
  //cout << "[replay LogWrite to " << file << "  bytes: " << size << "  offset: " << offset << "   data object: " << data << "]\n";
  Overlay *o = new Overlay(offset, size, data, 0, event_id);
  FileSummary *f = (FileSummary *) dfs->get_summary(file);

  f->getattr(&attr);
  
  if(attr.st_size < (off_t) (offset + size)) {
    attr.st_size = offset + size;
    f->setattr(&attr, event_id);    
  }

  f->add_overlay(o);

  END_REPLAY;
}

void DFS::read_segment(void *dest, segment_id i, segment_offset offset, segment_offset size) {
  map<segment_id,frame_id>::iterator itr = segment_logical_to_physical.find(i);
  assert(itr != segment_logical_to_physical.end());
  frame_id frame = itr->second;
  
  if(frame == nframes) {
    pthread_mutex_lock(&pending_lock);    
    // segment is in open segments, but we need to be careful with the locks so that pending commit doesn't give us 
    // an inconsistent open_segments
    for(list<Segment *>::iterator j = open_segments.begin(); j != open_segments.end(); j++) {
      Segment *s = *j;
      if(s->header->id == i) {
	// we've found our open segment
	assert(offset < SEGMENT_SIZE_BYTES && offset+size <= SEGMENT_SIZE_BYTES);
	memcpy(dest, &(s->data[offset]), size);
	pthread_mutex_unlock(&pending_lock);   
	return;
      }
    }
    pthread_mutex_unlock(&pending_lock);    
    assert(false); // fail, didn't find open segment in open segments list!!
  } else {
    read_frame(dest, frame, offset, size);
  }
}


void DirectorySummary::set_entry(string name, inode_id i, object_id creator) {
  map<string,inode_id>::iterator j = dirent_diff.find(name);
  if(j != dirent_diff.end()) { 
    // unlink old entry
    inode_id old = j->second;
    dirent_diff.erase(j);
    if(old != DELETED) {
      Summary *old_summary = dfs->get_summary(old);
      struct stat oldattr;
      old_summary->getattr(&oldattr);
      if(!oldattr.st_mode & S_IFDIR) {
	oldattr.st_nlink--;
	old_summary->setattr(&oldattr, creator);
	if(oldattr.st_nlink <= 0) 
	  dfs->delete_summary(old);
      }
    }
  }
  dirent_diff[name] = i;
}

// fix later!
#define CacheOverlay Overlay

void LogCache::replay() {
  BEGIN_REPLAY;

  struct stat attr;
  
  // cout << "[replay LogCache to " << file << "  bytes: " << size << "  offset: " << offset << "   data object: " << data << "]\n";
  CacheOverlay *o = new CacheOverlay(offset, size, data, 0, event_id);
  FileSummary *f = (FileSummary *) dfs->get_summary(file);

  f->getattr(&attr);

  
  if(attr.st_size < (off_t) (offset + size)) {
    // shouldn't happen
    if(offset > attr.st_size) {
      // pad with zeroes
      CacheOverlay *pad = CacheOverlay::zeroes(attr.st_size, offset-attr.st_size, event_id);
      f->add_overlay(pad);
    }
    attr.st_size = offset + size;
    f->setattr(&attr, event_id);    
  }

  f->add_overlay(o);

  END_REPLAY;
}

off_t FileSummary::read_base_chunk(void *buffer, off_t offset, size_t size) {

  cout << "FileSummary: READ BASE CHUNK " << base_path << "\n";
  // read chunk from base
  dfs->base->read(buffer, base_path, offset, size);

  DataObject *data = new DataObject(string((char *)buffer,size));
  (new LogCache(my_id, offset, size, dfs->append_data_object(data)))->play();

  return size;
}


void LogUsePOSIXBase::replay() {
  BEGIN_REPLAY;
  dfs->base = new POSIXBase(root_path);
  dfs->baptize(dfs->get_root_ino(), dfs->base->root_path);
  END_REPLAY;
}

void DFS::dfs_statfs(fuse_req_t req, fuse_ino_t ino) {

  struct statvfs stbuf;
  memset(&stbuf,0,sizeof(struct statvfs));
  
  int base_size = 0; //1024*1024*1024/8192;

  stbuf.f_bsize = 8192; // not really relevant
  stbuf.f_frsize = 8192;
  stbuf.f_blocks = base_size + nframes * SEGMENT_SIZE_BYTES / stbuf.f_frsize;
  stbuf.f_bfree = n_empty_frames() * SEGMENT_SIZE_BYTES / stbuf.f_frsize; // FIXME: NUMBER OF FREE BLOCKS
  stbuf.f_bavail = stbuf.f_bfree;
  stbuf.f_files = 1024; // FIXME NUMBER OF INODES
  stbuf.f_ffree = 1024; // FIXME NUMBER OF FREE INODES
  stbuf.f_favail = stbuf.f_ffree;
  stbuf.f_fsid = 0xbeef; // FIXME ?
  stbuf.f_flag = 0; // no mount flags
  stbuf.f_namemax = 512; // not a real limit...
  
  fuse_reply_statfs(req, &stbuf);
}


/* ------------- frame management functions -------------------------------*/
// get an empty segment frame... very stupid right now
frame_id DFS::empty_segment_frame() {
  if(frame_counter > nframes) 
    return NoFreeSegments;
  else
    // hack
    return ((frame_counter++)+zero_frame)%nframes;
}

unsigned int DFS::n_empty_frames() {
  return (nframes - frame_counter + 1);
}
/*--------------------------------------------------*/

void DFS::crawl() {
  cout << " crawling...\n";

  map<ino_t, inode_id>::iterator i = inode_map.find(get_root_ino());
  assert(i != inode_map.end());
  
  DirectorySummary *s = (DirectorySummary *) get_summary(i->second,false);
  assert(s != NULL);

  crawl(s);
  cout << " done.\n";
}

void DFS::crawl(DirectorySummary *s) {
  s->fill_dirent_diff();
  map<string,inode_id>::iterator i;
  for(i = s->dirent_diff.begin(); i != s->dirent_diff.end(); i++) {
    if(i->first == string(".") || i->first == string(".."))
      continue;
    inode_id next_inode = i->second;
    if(next_inode == DELETED || next_inode == 0) 
      continue;
    Summary *next = get_summary(next_inode, false);
    assert(next != NULL);
    if(next->type == DIRECTORY_SUMMARY) {
      crawl((DirectorySummary *)next);
    }
  }
}

void DFS::freeze() { 
  Event *newbase = base->freeze();
  if(newbase != NULL) {
    cout << "Switching to new base\n";
    Event *zero = new LogZero(empty_segment_frame());
    delete base;
    base = NULL;
    zero->play();
    newbase->play();
    assert(base != NULL);
    baptize(get_root_ino(), base->root_path);
    cout << "Committing final freeze log\n";
    commit_pending_log(true);
    cout << "  Done.\n";
  }
}

void DFS::nuke() {
  pending_log_entries.clear();
  open_segments.clear();
  pending_data_entries.clear();
  for(map<inode_id,Summary*>::iterator i = summaries.begin();
      i != summaries.end(); i++) {
    Summary *s = i->second;
    delete s;
  }
  summaries.clear();
  segment_logical_to_physical.clear();
  inode_map.clear();
  object_info.clear();
  open_segments.clear(); // memory leak?
  inode_id_counter = 0;
}

void LogZero::replay() {
  BEGIN_REPLAY;
  dfs->zero_frame = zero_frame;
  dfs->frame_counter = 0;
  END_REPLAY;
}

string LogSetattr::compute_data() {
  return serialize_to_string( (const LogSetattr &) *this);
}

string LogCreateRegularFile::compute_data() {
  return serialize_to_string( (const LogCreateRegularFile &) *this);
}

string LogSetDirent::compute_data() {
  return serialize_to_string( (const LogSetDirent &) *this);
}

string LogWrite::compute_data() {
  return serialize_to_string( (const LogWrite &) *this);
}

string LogUsePOSIXBase::compute_data() {
  return serialize_to_string( (const LogUsePOSIXBase &) *this);
}

string LogMapINode::compute_data() {
  return serialize_to_string( (const LogMapINode &) *this);
}

string LogCreateDirectory::compute_data() {
  return serialize_to_string( (const LogCreateDirectory &) *this);
}

string LogCache::compute_data() {
  return serialize_to_string( (const LogCache &) *this);
}

string LogUseS3ROBase::compute_data() {
  return serialize_to_string( (const LogUseS3ROBase &) *this);
}

string LogSetS3ROType::compute_data() {
  return serialize_to_string( (const LogSetS3ROType &) *this);
}

string LogZero::compute_data() {
  return serialize_to_string( (const LogZero &) *this);
}

template <class T>
string serialize_to_string(T &x) {
  ostringstream ostrm(ios_base::out | ios_base::binary);
  boost::archive::binary_oarchive oa(ostrm);
  oa << x;
  return ostrm.str();
}

void dirbuf_add(fuse_req_t req, struct dirbuf *b, const char *name,
		fuse_ino_t ino)
{
  struct stat stbuf;
  size_t oldsize = b->size;
  b->size += fuse_add_direntry(req, NULL, 0, name, NULL, 0);
  char *newp = (char *)realloc(b->p, b->size);
  if (!newp)
    {
      fprintf(stderr, "*** fatal error: cannot allocate memory\n");
      abort();
    }
  b->p = newp;
  memset(&stbuf, 0, sizeof(stbuf));
  stbuf.st_ino = ino;
  fuse_add_direntry(req, b->p + oldsize, b->size - oldsize, name, &stbuf,
		    b->size);
}

int reply_buf_limited(fuse_req_t req, const char *buf, size_t bufsize,
			     off_t off, size_t maxsize)
{
  if ((size_t) off < bufsize)
    return fuse_reply_buf(req, buf + off, std::min(bufsize - (size_t) off, maxsize));
  else
    return fuse_reply_buf(req, NULL, 0);
}

