#ifndef _INCLUDE_DFS
#define _INCLUDE_DFS

#include <time.h>
#include <pthread.h>
#include <fstream>
#include <sstream>
#include <string>
#include <list>
#include <vector>
#include <algorithm>
#include <map>
#include <sfs/async.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <dirent.h>

#include <fcntl.h>
#include <openssl/sha.h>

#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/serialization/list.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/map.hpp>
#include <boost/serialization/base_object.hpp>

#define FUSE_USE_VERSION 26
#include <fuse_lowlevel.h>

#include "base.hh"

using namespace std;

//#define BOOST_ARCHIVE_ITYPE boost::archive::text_iarchive
//#define BOOST_ARCHIVE_OTYPE boost::archive::text_oarchive
#define BOOST_ARCHIVE_ITYPE boost::archive::binary_iarchive
#define BOOST_ARCHIVE_OTYPE boost::archive::binary_oarchive

//#define BEGIN_REPLAY dfs->begin_replay()
//#define END_REPLAY dfs->end_replay()

#define BEGIN_REPLAY (cout << "Replay " << *this << "\n")
#define END_REPLAY

//#define P dfs->p()
//#define V dfs->v()

#define P
#define V

#include "dfs_types.h"

#include "log.hh"

#define PERCENT(x,y) ((int)((double)(x) / (double)(y) * 100))

struct dirbuf
{
	char *p;
	size_t size;
};

void dirbuf_add(fuse_req_t req, struct dirbuf *b, const char *name,
		fuse_ino_t ino);


int reply_buf_limited(fuse_req_t req, const char *buf, size_t bufsize,
		      off_t off, size_t maxsize);


/*----------------------------------------------------------*/

/* Log operations */

// Note: Paths relative to fuse/base root




// Associate dfs_id with ino_t


// Associate dirent lookup path <dfs_id,name> with ino_t

// Associate root ino_t and dfs_id

// Create new regular file with dfs_id

// Create new directory with dfs_id

// Create dir entry (hard link)  <dfs_id,name> -> dfs_id

// Remove dir entry (unlink) <dfs_id,name> 

// Create new symbolic link <dfs_id,path>



struct event_summary_comparison {
  template <typename T>
  bool operator() (T &a, T &b) {
    return a.id < b.id;
  }
};

const summary_type FILE_SUMMARY = 1;
const summary_type DIRECTORY_SUMMARY = 2;
const summary_type STUB_SUMMARY = 3;

void verify_assumptions();

class Overlay {
 public:
  object_id data;
  long data_offset;
  long offset;
  long size;
  object_id creator;

  Overlay(long o, long s, object_id d, long d_offset, object_id creator) {
    data = d;
    offset = o;
    size = s;
    data_offset = d_offset;
    this->creator = creator;
  }

  Overlay(long o, long s, object_id creator) {
    data = 0;
    offset = o;
    size = s;
    data_offset = 0;
    this->creator = creator;
  }

  Overlay(Overlay *o) {
    data = o->data;
    offset = o->offset;
    size = o->size;
    data_offset = o->data_offset;
    creator = o->creator;
  }

  static Overlay *zeroes(long offset, long size, object_id creator) {
    return new Overlay(offset,size,creator);
  }

  long extent() {
    return offset + size;
  }

  void ftruncate(long new_offset) { 
    assert(new_offset > offset);
    long diff = new_offset - offset;
    size -= diff;
    data_offset += diff;
    offset = new_offset;
  }

  void truncate(long new_extent) {
    assert(new_extent > offset);
    size = new_extent - offset;
  }

  friend bool operator<(const Overlay& x, const Overlay& y) {
    return (x.offset < y.offset);
  };

};

class Summary {
 public:
  string base_path;
  struct stat *mstat;
  object_id mstat_creator;
  object_id creator;
  summary_type type;
  inode_id my_id;
  bool in_base;

  Summary(inode_id me, string base, summary_type t, object_id creator) { 
    my_id = me;
    type = t; 
    this->creator = creator;
    mstat = NULL;
    mstat_creator = 0;
    base_path = base;
    in_base = true;
  }

  Summary(inode_id me, struct stat *mmstat, summary_type t, object_id creator) { 
    my_id = me;
    type = t; 
    mstat_creator = creator;
    this->creator = creator;
    assert(mmstat != NULL);
    mstat = new struct stat;
    memcpy(mstat, mmstat, sizeof(struct stat));
    in_base = false;
  }

  virtual void getattr(struct stat *attr);
  virtual void setattr(struct stat *attr, object_id creator);

  ~Summary() {
    if(mstat) 
      delete mstat;
  }

};

class StubSummary : public Summary {
 public:
  StubSummary(inode_id me, string base, object_id creator) : Summary(me, base, STUB_SUMMARY, creator) {}

  Summary *complete();
};

class FileSummary : public Summary {
public:
  vector<Overlay *> content_diff; // kept sorted by offset
  FileSummary(inode_id me, string base, object_id creator) : Summary(me, base, FILE_SUMMARY, creator) {}
  
  FileSummary(inode_id me, struct stat *attr, object_id creator) : Summary(me, attr, FILE_SUMMARY, creator) {}
  
  virtual void setattr(struct stat *attr, object_id creator);

  void change_size(size_t oldsize, size_t newsize, object_id creator);

  void add_overlay(Overlay *o);
  
  size_t read(char *buffer, offset_t offset, size_t size);
  
  off_t read_base_chunk(void *buffer, off_t offset, size_t size);


  void dump_content() {
    int j = 0;
    for(vector<Overlay *>::iterator i = content_diff.begin(); i != content_diff.end(); i++) {
      Overlay *o = *i;
      cout << "["<<j<<"] offset " << o->offset <<"-" << (o->extent()-1) << " size " << o->size << "   data " << o->data << "  data offset " << o->data_offset << "\n";
      j++;
    }
  }

  ~FileSummary() { 
    for(unsigned int i = 0; i < content_diff.size(); i++) {
      delete content_diff[i];
      content_diff[i] = NULL;
    }
  }

};

const inode_id DELETED = 0;

class DirectorySummary : public Summary {
 private:
  bool dirent_diff_filled;

 public:
  map<string,inode_id> dirent_diff;

  DirectorySummary(inode_id me, string base, object_id creator) : Summary(me, base, DIRECTORY_SUMMARY, creator) {
    dirent_diff_filled = false;
  }
  DirectorySummary(inode_id me, struct stat *attr, object_id creator) : Summary(me, attr, DIRECTORY_SUMMARY, creator) {
    dirent_diff_filled = false;
  }
  
  void fill_dirent_diff();

  void delete_entry(string name, object_id creator) {
    set_entry(name, DELETED, creator);
  }
  
  void directory_stream(fuse_req_t req, struct dirbuf &b);
  void set_entry(string name, inode_id i, object_id creator);

  bool empty();

  bool base_empty();
  
  inode_id lookup(string name);

  inode_id base_lookup(string name);

}; 

class DFS {
 private:



  inode_id inode_id_counter;
  segment_id segment_sequence_counter; // max known sequence number + 1
  object_id log_sequence_counter;

  pthread_mutex_t replay_lock;
  pthread_mutex_t fuse_lock;

 public:
  frame_id frame_counter;
  pthread_mutex_t pending_lock;
  uint32 nframes;
  uint32 zero_frame;
  map<object_id, event_summary> object_info;
  // maps base inodes to dfs inodes
  map<inode_id, Summary *> summaries;
  map<segment_id, frame_id> segment_logical_to_physical;
  map<ino_t, inode_id> inode_map;

  list<Event *> pending_log_entries;
  list<Segment *> open_segments;
  list<DataObject *> pending_data_entries;
  Base *base; 
  string block_dev;

  void freeze();
  void nuke();

  Summary *get_summary(inode_id i, bool assert_fail);
  Summary *get_summary(inode_id i);

  void delete_summary(inode_id i);
  void read_volume();
  void crawl();
  void crawl(DirectorySummary *s);

  void read_frame(ifstream &ifs, frame_id i, segment_offset offset) {
    ifs.open(block_dev.c_str(), ios_base::in | ios_base::binary);
    ifs.seekg(frame_offset(i)+offset);
  }

  void read_segment(void *dest, segment_id i, segment_offset offset, segment_offset bytes);

  void read_frame(void *dest, frame_id i, segment_offset offset, segment_offset bytes) {
    FILE *f = fopen(block_dev.c_str(),"r"); 
    assert(fseek(f, frame_offset(i)+offset, SEEK_SET) == 0);
    assert(fread(dest, bytes, 1, f) == 1);
    fclose(f);
  }

  void begin_replay() {
    pthread_mutex_lock(&replay_lock);
  }

  void end_replay() {
    pthread_mutex_lock(&replay_lock);
  }

  void p() {
    pthread_mutex_lock(&fuse_lock);
  }

  void v() {
    pthread_mutex_lock(&fuse_lock);
  }


  ino_t get_root_ino() {
    struct stat st;
    assert(base->get_stat(base->root_path,&st) == 0);
    return st.st_ino;    
  }

  DirectorySummary *get_root_summary() {
    map<ino_t, inode_id>::iterator i;
    i = inode_map.find(get_root_ino());
    assert(i != inode_map.end());
    return (DirectorySummary *)get_summary(i->second);
  }

  inode_id baptize(ino_t base_inode, string base_path); 

  void replay_log(event_summary &summary);

  template <class T>
    void replay_serialized(event_summary &summary);

  void write_frame(void *src, frame_id i, segment_offset offset, segment_offset bytes) {
    FILE *f = fopen(block_dev.c_str(),"r+"); 
    assert(fseek(f, frame_offset(i)+offset, SEEK_SET) == 0);
    assert(fwrite(src, bytes, 1, f) == 1);
    fclose(f);
  }



  static long frame_offset(frame_id i) {
    return i << SEGMENT_SIZE_BITS;
  } 
 
  void write_frame(void *src, frame_id i) {
    // write entire frame
    write_frame(src, i, 0, SEGMENT_SIZE_BYTES);
  }

  void seek_frame(frame_id i, FILE *f) {
    assert(fseek(f, frame_offset(i), SEEK_SET) == 0);
  }
  
  segment_id next_segment_id() {
    return ++segment_sequence_counter;
  }
  
  object_id next_object_id() {
    return ++log_sequence_counter;
  }

  inode_id next_inode_id() {
    //    cout << "NEXT ID COUNTER " << inode_id_counter << "\n";
    return ++inode_id_counter;
  }

  DFS(string volume);

  //  DFS(const char *block_dev_path, const char *base_path);   

  void use_inode_id(inode_id i) {
    inode_id_counter = std::max(inode_id_counter,i);
  }

  void destroy() {
    // commit pending events
    cout << "Shuttind down DFS: Commit all pending events.\n";
    commit_pending_log(true);

    // delete allocated inode summaries
    for(map<inode_id,Summary *>::iterator i = summaries.begin(); i != summaries.end(); i++) {
      delete i->second;
    }
  }

  frame_id empty_segment_frame();
  
  unsigned int n_empty_frames();
  
  static int format_volume(string volume, Event *base);

  static int format_volume(string volume, Event *base, bool wipe);

  void dfs_statfs(fuse_req_t req, fuse_ino_t ino);

  /* man 2 path_resolution */
  void dfs_lookup(fuse_req_t req, fuse_ino_t parent, const char *name);

  /* man 2 stat ( also check http://en.wikipedia.org/wiki/Stat_(Unix) ) */
  void dfs_getattr(fuse_req_t req, fuse_ino_t ino, struct fuse_file_info *fi);

  /* man 2 stat, man 2 chmod */
  void dfs_setattr(fuse_req_t req, fuse_ino_t ino, struct stat *attr,
		   int to_set, struct fuse_file_info *fi);

  /* man 2 mknod */
  void dfs_mknod(fuse_req_t req, fuse_ino_t parent, const char *name,
		 mode_t mode, dev_t rdev);

  /* man 2 mkdir */
  void dfs_mkdir(fuse_req_t req, fuse_ino_t parent, const char *name,
		 mode_t mode);

  /* man 2 unlink */
  void dfs_unlink(fuse_req_t req, fuse_ino_t parent, const char *name);
  
  /* man 2 rmdir */
  void dfs_rmdir(fuse_req_t req, fuse_ino_t parent, const char *name);

  /* man 2 rename --- annoyingly intricate
   * in principle, like unlink/rmdir + mknod/mkdir modulo corner cases Fuse 
   * takes care of the fact that newparent may not be a subdirectory of parent.
   */
  void dfs_rename(fuse_req_t req, fuse_ino_t parent, const char *name,
		  fuse_ino_t newparent, const char *newname);

  /* man 2 open -- do not use if filesystem is stateless */
  //void dfs_open(fuse_req_t req, fuse_ino_t ino, struct fuse_file_info *fi);

  /* man 2 close -- do not use if filesystem is stateless */
  //void dfs_release(fuse_req_t req, fuse_ino_t ino, struct fuse_file_info *fi);

  /* man 2 read */
  void dfs_read(fuse_req_t req, fuse_ino_t ino, size_t size, off_t off,
		struct fuse_file_info *fi);

  /* man 2 write, also check ``man 2 lseek'' for out of bounds write */
  void dfs_write(fuse_req_t req, fuse_ino_t ino, const char *buf,
		 size_t size, off_t off, struct fuse_file_info *fi);

  /* man 2 readdir */
  void dfs_readdir(fuse_req_t req, fuse_ino_t ino, size_t size, off_t off,
		   struct fuse_file_info *fi);

  void dfs_destroy(void *userdata);

  static bool fs_exists(string file) {
    struct stat st;
    if(stat(file.c_str(), &st))
      return errno != ENOENT;
    else 
      return true;
  }

  static off_t fs_get_size(string path) {
    struct stat st;
    if(stat(path.c_str(),&st)) {
      cerr << "stat error in get_size " << errno << "\n";
      return 0;
    }
    return st.st_size;
  }

  object_id append_log_object(Event *o) {
    pthread_mutex_lock(&pending_lock);
    o->event_id = next_object_id();
    pending_log_entries.push_back(o);
    pthread_mutex_unlock(&pending_lock);
    //commit_pending_log(false);
    return o->event_id;


  }

  object_id append_data_object(DataObject *o) {
    pthread_mutex_lock(&pending_lock);
    object_id oid = next_object_id();
    o->event_id = oid;
    pending_data_entries.push_back(o);
    pthread_mutex_unlock(&pending_lock);
    commit_pending_log(false);
    return oid;
  }

  void commit_pending_log() { return commit_pending_log(false); }
  void commit_pending_log(bool write_all_segments);

  void select_open_segment(uint32 needed_space, list<Segment*>::iterator &i) {
    assert(needed_space < MAX_SEGMENT_STORAGE);
    for(i = open_segments.begin() ; i != open_segments.end(); i++)
      if((*i)->free_space >= needed_space) return;
    // all segments are too full
    Segment *s = new Segment(next_segment_id());
    segment_logical_to_physical[s->header->id] = nframes;
    open_segments.push_front(s);
    assert(s->free_space >= needed_space);
    i = open_segments.begin();
  }


};



struct OverlayPointerComparison {
  template <typename T>
  bool operator() (T *a, T *b) {
    return a->offset < b->offset;
  }
};









#endif
