
#ifndef _LOG_INCLUDED_
#define _LOG_INCLUDED_

#include <string>
#include <iostream>

using namespace std;

// Log entry types
const entry_type LOG_TEST                   = 0xdead;
const entry_type LOG_TEST_DATA              = 0xbeef;
const entry_type LOG_ERROR                  = 0;
const entry_type LOG_DATA                   = 1;
const entry_type LOG_MAP_INODE              = 2;
const entry_type LOG_SETATTR                = 3;
const entry_type LOG_CREATE_REGULAR_FILE    = 4;
const entry_type LOG_SET_DIRENT             = 5;
const entry_type LOG_CREATE_DIRECTORY       = 6;
const entry_type LOG_WRITE                  = 7;
const entry_type LOG_CACHE                  = 8;
const entry_type LOG_POSIX_BASE             = 9;
const entry_type LOG_S3RO_BASE              = 10;
const entry_type LOG_S3RO_TYPE              = 11;
const entry_type LOG_ZERO                   = 12;

typedef off_t offset_t;

static string eventname(entry_type e) {
  switch(e) { 
  case LOG_TEST:                   return string("Test");
  case LOG_TEST_DATA:              return string("TestData");
  case LOG_ERROR:                  return string("Error");
  case LOG_DATA:                   return string("Data");
  case LOG_MAP_INODE:              return string("MapINode");
  case LOG_SETATTR:                return string("SetAttr");
  case LOG_CREATE_REGULAR_FILE:    return string("CreateRegularFile");
  case LOG_SET_DIRENT:             return string("SetDirent");
  case LOG_CREATE_DIRECTORY:       return string("CreateDirectory");
  case LOG_WRITE:                  return string("Write"); 
  case LOG_CACHE:                  return string("Cache");
  case LOG_POSIX_BASE:             return string("UsePOSIXBase");
  case LOG_S3RO_BASE:              return string("UseS3ROBase");
  case LOG_S3RO_TYPE:              return string("S3ROType");
  case LOG_ZERO:                   return string("Zero");
  default:
    cerr << "invalid log event type\n";
    assert(false);
    return string();
  }
}

struct event_summary {
  object_id id;
  segment_id segment;
  segment_offset offset;
  entry_type type;
  segment_offset size;
};


// segment constants

const uint32 NoFreeSegments                 = 0xffffffff;

#define SEGMENT_FRESHNESS_TIMEOUT      30 // seconds
#define SEGMENT_SIZE_BITS              17  // 128 KB
#define SEGMENT_SIZE_BYTES             (1 << SEGMENT_SIZE_BITS)
#define MAX_SEGMENT_STORAGE            (SEGMENT_SIZE_BYTES - sizeof(segment_header) - sizeof(segment_header_entry))
#define SEGMENT_ADDRESS_BYTES          3  // 3 byte segment addresses

// max filesystem size = 2**MAX_VOLUME_SIZE bytes
// currently 2 TB
#define MAX_VOLUME_SIZE                (SEGMENT_SIZE_BITS+(SEGMENT_ADDRESS_BYTES<<3))


#define PTR_ADD(p,x)                   ((void *)(((unsigned long)(p))+((unsigned long)(x))))

struct segment_header_entry {
  object_id id;
  segment_offset offset;
  segment_offset size;
  entry_type type;
};



class Event {
 private:
  string data_cache;
  bool have_cache;
  virtual void cache_data() {
    if(have_cache) return;
    have_cache = true;
    data_cache = compute_data();
  }

 public:
  object_id event_id;
  entry_type type;

  Event(entry_type t) { 
    have_cache = false; 
    event_id = 0;
    set_type(t);
  }

  void set_type(entry_type t) { type = t; }

  friend ostream &operator<<(ostream &stream, Event &o) {
    string r = o.repr();
    if(r.size() > 0) 
      r = string(" ") + r;

    stream << "<" << o.event_id << ": " << eventname(o.type) << o.repr() << ">";
    return stream;
  }

  virtual string repr() { return string();}

  virtual uint32 get_size() {
    return get_data().size();
  }

  virtual string get_data() { 
    cache_data();
    return data_cache;
  }

  virtual void invalidate_cache() {
    have_cache = false;
  }

  virtual string compute_data() {
    assert(false); // must be overridden
  }

  virtual entry_type get_type() { 
    return type;
  }
  
  virtual void replay() {}

  virtual void play();
};




bool is_log_event_type(entry_type t);

class DataObject : public Event {
 private:
  bool summary_cached;
  event_summary summary; // for reading
  
  void load_summary();
  
 public:
  string data;  // for writing
  object_id next;  // for writing
  object_id my_id; // for reading
 DataObject(string dat) : Event(LOG_DATA) {
    data = dat;
    next = 0;
    my_id = 0; // my_id is for reading
  }
  

  
 DataObject(object_id i) : Event(LOG_DATA) { 
    my_id = i;
    summary_cached = false;
  }

  size_t read(void *buffer, offset_t offset, size_t size);

  object_id get_next();

  frame_id get_frame(); 

  virtual string compute_data() {
    return string( (const char *) &next, sizeof(object_id)) + data;
  }

  DataObject *split(uint32 size, object_id new_id) {
    cout << "    split, size " << get_size() << "  to " << size << "\n";
    assert(size > sizeof(object_id));
    cout << "      invalidate cache\n"; flush(cout);
    invalidate_cache();

    uint32 local = size - sizeof(object_id);

    cout << "      slice substring\n"; flush(cout);
    size_t ds = data.size();
    cout << "   original size " << ds;
    flush(cout);
    cout << "   target size " << local;
    flush(cout);
    cout << "   slice " << local << ", " << (data.size()-local);
    flush(cout);

    cout << "\n"; flush(cout);

    string subst = data.substr(local,data.size()-local);
    
    cout << "      create new object\n"; flush(cout);
    DataObject *n = new DataObject(subst);
    cout << "      truncate\n"; flush(cout);
    data = data.substr(0,local);
    assert(get_size() == size);
    return n;
  }
};

struct segment_header {
  segment_id id;
  nobject_type n_objects;
};

class Segment {
 private:


  segment_offset pos;

  // for reading from disk
  bool header_read;
  bool data_read;
  bool table_read;
  frame_id frame;

 public:
  unsigned char data[SEGMENT_SIZE_BYTES];
  segment_header_entry *table;
  segment_header *header;
  uint32 free_space;
  time_t creation;
  Segment(segment_id i) {
    creation = time(NULL);
    memset(data,0,SEGMENT_SIZE_BYTES);
    free_space = SEGMENT_SIZE_BYTES - \
      sizeof(segment_header_entry) - \
      sizeof(segment_header);
    header = (segment_header *) &(data[SEGMENT_SIZE_BYTES - sizeof(segment_header)]); //PTR_ADD(data,SEGMENT_SIZE_BYTES-sizeof(segment_header));

    header->n_objects = 0;
    header->id = i;


    table = (segment_header_entry *) PTR_ADD(header,-sizeof(segment_header_entry));

    pos = 0;
  }

  //  string fetch_object(unsigned int index);

  void read_table();


  Segment(frame_id fr, bool foo);


  uint32 get_free_space() {
    return free_space;
  }

  void append(Event *o) {
    event_summary foo;
    append(o,&foo);
  }

  void append(Event *o, event_summary *summary) {
    assert(o != NULL);
    string dat = o->get_data();
    uint32 ds = dat.size();    
    assert(ds <= free_space);

    summary->segment = header->id;

    table->id = o->event_id; // header->id;
    summary->id = o->event_id;
    table->offset = pos;
    summary->offset = table->offset;
    table->size = ds;
    summary->size = table->size;
    table->type = o->get_type();
    summary->type = table->type;

    memcpy(PTR_ADD(data,pos), dat.c_str(), ds);

    pos += ds;
    free_space -= ds;
    if(free_space < sizeof(segment_header_entry))
      free_space = 0;
    else
      free_space -= sizeof(segment_header_entry);

    assert(free_space <= SEGMENT_SIZE_BYTES);

    table = (segment_header_entry *) PTR_ADD(table, -sizeof(segment_header_entry)); 
    header->n_objects += 1;
    delete o;
  }

  bool full() {
    return (free_space <= 512);
  }

  int write(frame_id fr);

};




#include "log_objects.hh"

#endif
