#ifndef LOG_OBJECTS_INCLUDED
#define LOG_OBJECTS_INCLUDED

#include "serialize_to_string.hh"

#include <openssl/sha.h>

static string sha1(string data) {
  SHA_CTX sha1_context;
  SHA1_Init(&sha1_context);
  SHA1_Update(&sha1_context, data.c_str(), data.length());
  unsigned char buffer[SHA_DIGEST_LENGTH];
  SHA1_Final(buffer, &sha1_context);
  return string( (const char *) buffer, SHA_DIGEST_LENGTH);
  //  return base16(buffer, SHA_DIGEST_LENGTH);
}



class TestEntry : public Event {
public:
  string test;
  TestEntry(string teststring) : Event(LOG_TEST) {
    test = teststring;
  }

  virtual string compute_data() {
    return test;
  }


};



class TestDataEntry : public Event {
public:
  string hash;
  object_id data_ptr;
  TestDataEntry(DataObject *o) : Event(LOG_TEST_DATA) {
    assert(o->event_id != 0);
    data_ptr = o->event_id;
    hash = sha1(o->data);
  }


  virtual string compute_data() {
    return hash;
  }
};


class LogSetattr : public Event {
 private:
  friend class boost::serialization::access;
  template<class Archive>
    void serialize(Archive &ar, const unsigned int version) 
    {
      ar & ino;
      ar & attr;
    }
 public:
  string attr;
  inode_id ino;
 LogSetattr() : Event(LOG_SETATTR) {}
 LogSetattr(inode_id i, struct stat &at) : Event(LOG_SETATTR) {
    ino = i; 
    attr = string((char *)&at, sizeof(struct stat)); // hack so we can serialize it
  }

  virtual string compute_data();

  void replay();
};

class LogCreateRegularFile : public Event {
 private:
  friend class boost::serialization::access;
  template<class Archive>
    void serialize(Archive &ar, const unsigned int version) 
    {
      ar & file_id;
      ar & attr_string;
    }
 public:
  inode_id file_id;
  string attr_string;
  
  string repr() { 
    ostringstream temp;
    temp << " file: " << file_id;
    return temp.str();
  }
 LogCreateRegularFile() : Event(LOG_CREATE_REGULAR_FILE) {}
 LogCreateRegularFile(inode_id iid, struct stat &attr) : Event(LOG_CREATE_REGULAR_FILE) {
    file_id = iid;
    attr_string = string((char *)&attr, sizeof(struct stat)); // hack so we can serialize it    
  }


  void getattr(struct stat &attr) { 
    memcpy(&attr, attr_string.c_str(), sizeof(struct stat));
  }

  virtual string compute_data();
  
  virtual void replay();
}; 

class LogSetDirent : public Event {
 private:
  friend class boost::serialization::access;
  template<class Archive>
    void serialize(Archive &ar, const unsigned int version) 
    {
      ar & name;
      ar & parent;
      ar & id;
    }
  
 public:

  string name;
  inode_id parent;
  inode_id id;

  string repr() { 
    ostringstream temp;
    if(id == 0) // DELETED
      temp << "parent " << parent << " DELETE " << name;
    else
      temp << "parent " << parent << " id " << id << " name " << name;

    return temp.str();
  }

 LogSetDirent() : Event(LOG_SET_DIRENT) {}
 LogSetDirent(string n, inode_id p, inode_id i) : Event(LOG_SET_DIRENT) {
    name = n;
    id = i;
    parent = p;
  }
  
  virtual string compute_data();
  void replay();

};

class LogWrite : public Event {
 private:
  friend class boost::serialization::access;
  template<class Archive>
    void serialize(Archive &ar, const unsigned int version) 
    {
      ar & file; ar & offset; ar & size; ar & data;
    }
 public:
  object_id file;
  offset_t offset;
  size_t size;
  object_id data;
 LogWrite() : Event(LOG_WRITE) {}


 LogWrite(object_id file, offset_t offset, size_t size, object_id data) : Event(LOG_WRITE) {
    this->file = file; this->offset = offset; this->size = size; this->data = data;
  }

  virtual string compute_data();
  
  virtual void replay();

  virtual string repr() {
    ostringstream temp;
    temp << "file: " << file << " offset: " << offset << " size: " << size << " data: " << data;
    return temp.str();
  }

};

class LogUsePOSIXBase : public Event {
private:
  friend class boost::serialization::access;
  template<class Archive>
    void serialize(Archive &ar, const unsigned int version) 
    {
      ar & root_path;
    }
  
 public:
  string root_path;

  string repr() {
    ostringstream temp;
    temp << " root_path " << root_path;
    return temp.str();
  }

  LogUsePOSIXBase() : Event(LOG_POSIX_BASE) {}
  
  LogUsePOSIXBase(string root_path) : Event(LOG_POSIX_BASE) {
    this->root_path = root_path;
  }
  
  virtual string compute_data();

  void replay();
};

class LogMapINode : public Event {

 private:
  friend class boost::serialization::access;
  template<class Archive>
    void serialize(Archive &ar, const unsigned int version) 
  {
      ar & base_ino;
      ar & id;
      ar & base_path;
    }
  
 public:
  ino_t base_ino;
  inode_id id;
  string base_path;

  string repr() {
    ostringstream temp;
    temp << " inode " << id << " = " << base_path;
    return temp.str();
  }

 LogMapINode() : Event(LOG_MAP_INODE) {}

 LogMapINode(ino_t base_ino, inode_id dfs_id, string base) : Event(LOG_MAP_INODE) {
    this->base_ino = base_ino;
    id = dfs_id;
    base_path = base;
  }
  
  
  virtual string compute_data();
  void replay();


};

class LogCreateDirectory : public Event {
 private:
  friend class boost::serialization::access;
  template<class Archive>
    void serialize(Archive &ar, const unsigned int version) 
    {
      ar & directory_id;
      ar & attr_string;
    }
 public:
  inode_id directory_id;
  string attr_string;
 LogCreateDirectory() : Event(LOG_CREATE_DIRECTORY) {}
 LogCreateDirectory(inode_id iid, struct stat &attr) : Event(LOG_CREATE_DIRECTORY) {
    directory_id = iid;
    attr_string = string((char *)&attr, sizeof(struct stat)); // hack so we can serialize it    
  }

  void getattr(struct stat &attr) { 
    memcpy(&attr, attr_string.c_str(), sizeof(struct stat));
  }

  virtual string compute_data();
  
  virtual void replay();
}; 


class LogCache : public Event {
 private:
  friend class boost::serialization::access;
  template<class Archive>
    void serialize(Archive &ar, const unsigned int version) 
    {
      ar & file; ar & offset; ar & size; ar & data;
    }
 public:
  object_id file;
  offset_t offset;
  size_t size;
  object_id data;
 LogCache() : Event(LOG_CACHE) {}

 LogCache(object_id file, offset_t offset, size_t size, object_id data) : Event(LOG_CACHE) {
    this->file = file; this->offset = offset; this->size = size; this->data = data;
  }

  virtual string compute_data();
  
  virtual void replay();

};


class LogUseS3ROBase : public Event {
private:
  friend class boost::serialization::access;
  template<class Archive>
    void serialize(Archive &ar, const unsigned int version) 
    {
      ar & key_id;
      ar & secret_key;
      ar & bucket;
      ar & root;
    }
  
 public:
  string key_id;
  string secret_key;
  string bucket;
  string root;

  string repr() {
    ostringstream temp;
    temp << " bucket " << bucket << " root " << root;
    return temp.str();
  }

  LogUseS3ROBase() : Event(LOG_S3RO_BASE) {}
  
  LogUseS3ROBase(string key_id, string secret_key, string bucket, string root) : Event(LOG_S3RO_BASE) {
    this->key_id = key_id;
    this->secret_key = secret_key;
    this->bucket = bucket;
    this->root = root;
  }
  
  virtual string compute_data();

  void replay();
};

class LogSetS3ROType : public Event {
private:
  friend class boost::serialization::access;
  template<class Archive>
    void serialize(Archive &ar, const unsigned int version) 
    {
      ar & path;
      ar & type;
    }
  
 public:
  string path;
  short type;

  string repr() {
    ostringstream temp;
    temp << " path " << path << " type " << type;
    return temp.str();
  }

  LogSetS3ROType() : Event(LOG_S3RO_TYPE) {}
  
  LogSetS3ROType(string path, short type) : Event(LOG_S3RO_TYPE) {
    this->path = path;
    this->type = type;
  }
  
  virtual string compute_data();
  void replay();
};

class LogZero : public Event {
private:
  friend class boost::serialization::access;
  template<class Archive>
    void serialize(Archive &ar, const unsigned int version) 
    {
      ar & zero_frame;
    }
  
 public:
  frame_id zero_frame;

  LogZero() : Event(LOG_ZERO) {}

  LogZero(frame_id zero_frame) : Event(LOG_ZERO) {
    this->zero_frame = zero_frame;
  }
  
  virtual string compute_data();
  void replay();
};





#endif
