/*
 * @author Tudor Marian <tudorm@cs.cornell.edu>
 */

#include "s3ops.h"

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <time.h>
#include <string.h>
#include <errno.h>

#include "list.h"

/* based on the s3.c in the libs3/src */

#define MAX_KEYS 0x7fffffff
#define min(a, b) (( (a) < (b) ) ? (a) : (b))
static S3Protocol protocolG = S3ProtocolHTTPS;
static S3UriStyle uriStyleG = S3UriStylePath;
static int retriesG = 5;
static int showResponsePropertiesG = 0;
static S3Status statusG = S3StatusOK;
static char errorDetailsG[4096] = { 0 };

static const char *accessKeyIdG = NULL;
static const char *secretAccessKeyG = NULL;

static int should_retry(void)
{
	if (retriesG--)
	{
		// Sleep before next retry; start out with a 1 second sleep
		static int retrySleepInterval = 1;
		sleep(retrySleepInterval);
		// Next sleep 1 second longer
		retrySleepInterval++;
		return 1;
	}

	return 0;
}

// This callback does the same thing for every request type: prints out the
// properties if the user has requested them to be so
static S3Status responsePropertiesCallback(
		const S3ResponseProperties *properties, void *callbackData)
{
	(void) callbackData;

	if (!showResponsePropertiesG)
	{
		return S3StatusOK;
	}

#define print_nonnull(name, field)                                 \
    do {                                                           \
        if (properties-> field) {                                  \
            printf("%s: %s\n", name, properties-> field);          \
        }                                                          \
    } while (0)

	print_nonnull("Content-Type", contentType);
	print_nonnull("Request-Id", requestId);
	print_nonnull("Request-Id-2", requestId2);
	if (properties->contentLength > 0)
	{
		printf("Content-Length: %lld\n",
				(unsigned long long) properties->contentLength);
	}
	print_nonnull("Server", server);
	print_nonnull("ETag", eTag);
	if (properties->lastModified > 0)
	{
		char timebuf[256];
		time_t t = (time_t) properties->lastModified;
		// gmtime is not thread-safe but we don't care here.
		strftime(timebuf, sizeof(timebuf), "%Y-%m-%dT%H:%M:%SZ", gmtime(&t));
		printf("Last-Modified: %s\n", timebuf);
	}
	int i;
	for (i = 0; i < properties->metaDataCount; i++)
	{
		printf("x-amz-meta-%s: %s\n", properties->metaData[i].name,
				properties->metaData[i].value);
	}

	return S3StatusOK;
}

static void responseCompleteCallback(S3Status status,
		const S3ErrorDetails *error, void *callbackData)
{
	(void) callbackData;

	statusG = status;
	// Compose the error details message now, although we might not use it.
	// Can't just save a pointer to [error] since it's not guaranteed to last
	// beyond this callback
	int len = 0;
	if (error && error->message)
	{
		len += snprintf(&(errorDetailsG[len]), sizeof(errorDetailsG) - len,
				"  Message: %s\n", error->message);
	}
	if (error && error->resource)
	{
		len += snprintf(&(errorDetailsG[len]), sizeof(errorDetailsG) - len,
				"  Resource: %s\n", error->resource);
	}
	if (error && error->furtherDetails)
	{
		len += snprintf(&(errorDetailsG[len]), sizeof(errorDetailsG) - len,
				"  Further Details: %s\n", error->furtherDetails);
	}
	if (error && error->extraDetailsCount)
	{
		len += snprintf(&(errorDetailsG[len]), sizeof(errorDetailsG) - len,
				"%s", "  Extra Details:\n");
		int i;
		for (i = 0; i < error->extraDetailsCount; i++)
		{
			len += snprintf(&(errorDetailsG[len]), sizeof(errorDetailsG) - len,
					"    %s: %s\n", error->extraDetails[i].name,
					error->extraDetails[i].value);
		}
	}
}

void s3_init_credentials(const char *access_key, const char *secret_key)
{
	accessKeyIdG = access_key;
	secretAccessKeyG = secret_key;
}

static void s3_init(void)
{
	S3Status status;
	if ((status = S3_initialize("s3", S3_INIT_ALL)) != S3StatusOK)
	{
		fprintf(stderr, "Failed to initialize libs3: %s\n", 
				S3_get_status_name(status));
		exit(-1);
	}
}

#define s3_deinit	S3_deinitialize

static void printError(void)
{
	if (statusG < S3StatusErrorAccessDenied)
	{
		fprintf(stderr, "\nERROR: %s\n", S3_get_status_name(statusG));
	}
	else
	{
		fprintf(stderr, "\nERROR: %s\n", S3_get_status_name(statusG));
		fprintf(stderr, "%s\n", errorDetailsG);
	}
}

int s3_rm_object(const char *bucket, const char *key)
{
	int ret = 1;
	S3BucketContext bucketContext = { bucket, protocolG, uriStyleG,
			accessKeyIdG, secretAccessKeyG };

	S3ResponseHandler responseHandler = { 0, &responseCompleteCallback };

	s3_init();

	do
	{
		S3_delete_object(&bucketContext, key, 0, &responseHandler, 0);
	} while (S3_status_is_retryable(statusG) && should_retry());

	if ((statusG != S3StatusOK) && (statusG != S3StatusErrorPreconditionFailed))
	{
		printError();
		ret = 0;
	}

	s3_deinit();
	return ret;
}

struct clear_bucket_data_callback
{
	char *bucket;
	int isTruncated;
};

S3Status listBucketCallback(int isTruncated, const char *nextMarker,
		int contentsCount, const S3ListBucketContent *contents,
		int commonPrefixesCount, const char **commonPrefixes,
		void *callbackData)
{
	int i;
	struct clear_bucket_data_callback *dc = 
		(struct clear_bucket_data_callback *) callbackData; 
	dc->isTruncated = isTruncated;
	
	for (i = 0; i < contentsCount; i++) 
	{
		const S3ListBucketContent *content = &(contents[i]);
		s3_rm_object(dc->bucket, content->key);
	}
	return S3StatusOK;
}

int s3_clear_bucket(char *bucket)
{
	int ret = 1;
	struct clear_bucket_data_callback cb;
	
	S3BucketContext bucketContext = { bucket, protocolG, uriStyleG,
			accessKeyIdG, secretAccessKeyG };

	S3ListBucketHandler listBucketHandler = { { &responsePropertiesCallback,
			&responseCompleteCallback }, &listBucketCallback };

	cb.bucket = bucket;
	
	s3_init();

	do
	{
		cb.isTruncated = 0;
		do
		{
			S3_list_bucket(&bucketContext, NULL, NULL, NULL, MAX_KEYS, NULL,
					&listBucketHandler, &cb);
		} while (S3_status_is_retryable(statusG) && should_retry());
		if (statusG != S3StatusOK)
			break;
	} while (cb.isTruncated);

	if (statusG != S3StatusOK)
	{
		ret = 0;
		printError();
	}

	s3_deinit();
	return ret;
}

/* keep a track of the bytes sent so far, and remaining bytes */
struct s3_put_callback_data
{
	uint64_t sent;
	uint64_t total;
	void *buffer;
};

static int putObjectDataCallback(int bufsz, char *buffer, void *data)
{
	uint64_t remaining;
	int bytes;
	struct s3_put_callback_data* cd = (struct s3_put_callback_data*) data;

	remaining = cd->total - cd->sent;
	if (remaining == 0)
		return remaining;

	bytes = bufsz;
	if ((uint64_t) bytes > remaining)
		bytes = (int) remaining;

	/* the libs3 interface is useless ... I need to do a copy here */
	memcpy(buffer, &(((char *) cd->buffer)[cd->sent]), bytes);
	cd->sent += bytes;
	return bytes;
}

/* S3 only supports full object overload, buf contains byte_count bytes */
ssize_t s3_put_object(const char *bucket, const char *key, void *buf,
		uint64_t byte_count)
{
	ssize_t ret;
	struct s3_put_callback_data data = { 0, byte_count, buf };

	S3BucketContext bucketContext = { bucket, protocolG, uriStyleG,
			accessKeyIdG, secretAccessKeyG };

	S3PutObjectHandler putObjectHandler = { { &responsePropertiesCallback,
			&responseCompleteCallback }, &putObjectDataCallback };

	s3_init();

	do
	{
		S3_put_object(&bucketContext, key, byte_count, NULL, NULL,
				&putObjectHandler, &data);
	} while (S3_status_is_retryable(statusG) && should_retry());

	if (statusG != S3StatusOK)
	{
		ret = -1;
		printError();
	}
	else
		ret = data.sent;
	s3_deinit();
	return ret;
}

struct s3_read_callback_data
{
	uint64_t bytes_read;
	void *buffer;
	uint64_t buffer_size;
};

static S3Status readObjectDataCallback(int bufsz, const char *buffer,
		void *data)
{
	int bytes_read = bufsz;
	S3Status ret = S3StatusOK;
	struct s3_read_callback_data *cd = (struct s3_read_callback_data *) data;
	uint64_t remaining = cd->buffer_size - cd->bytes_read;

	if ((uint64_t) bytes_read > remaining)
	{
		ret = S3StatusAbortedByCallback;
		bytes_read = (int) remaining;
	}

	memcpy(&(((char *) cd->buffer)[cd->bytes_read]), buffer, bytes_read);
	cd->bytes_read += bytes_read;

	return ret;
}

/* buf is provided by the caller, and should have at least byte_count bytes */
ssize_t s3_read_object(const char *bucket, const char *key, void *buf,
		uint64_t start_byte, uint64_t byte_count)
{
	ssize_t ret;
	struct s3_read_callback_data data = { 0, buf, byte_count };

	S3BucketContext bucketContext = { bucket, protocolG, uriStyleG,
			accessKeyIdG, secretAccessKeyG };

	S3GetObjectHandler getObjectHandler = { { &responsePropertiesCallback,
			&responseCompleteCallback }, &readObjectDataCallback };

	s3_init();

	do
	{
		S3_get_object(&bucketContext, key, NULL, start_byte, byte_count, 0,
				&getObjectHandler, &data);
	} while (S3_status_is_retryable(statusG) && should_retry());

	if (statusG != S3StatusOK)
	{
		printError();
		ret = -1;
	}
	else
		ret = data.bytes_read;
	s3_deinit();

	return ret;
}

struct s3_get_callback_data_buf
{
	struct list_head list; /* to chain it */
	void *buf;
	size_t bufsize;
};

struct s3_get_callback_data
{
	struct list_head bufs;
	size_t total_bytes;
};

static S3Status getObjectDataCallback(int bufsz, const char *buffer, void *data)
{
	S3Status ret = S3StatusOK;
	struct s3_get_callback_data *cd = (struct s3_get_callback_data *) data;
	struct s3_get_callback_data_buf *cb_buf;

	if (bufsz == 0)
		return ret;
	
	if (!(cb_buf = (struct s3_get_callback_data_buf *) malloc(
			sizeof(struct s3_get_callback_data_buf))))
	{
		fprintf(stderr, "Malloc error, %s\n", strerror(errno));
		exit(1);
	}
	if (!(cb_buf->buf = malloc(bufsz)))
	{
		fprintf(stderr, "Malloc error, %s\n", strerror(errno));
		exit(1);
	}
	cb_buf->bufsize = bufsz;
	memcpy(cb_buf->buf, buffer, bufsz);

	/* maintain accounting */
	list_add_tail(&cb_buf->list, &cd->bufs);
	cd->total_bytes += cb_buf->bufsize;

	return ret;
}

ssize_t s3_get_object(const char *bucket, const char *key, void **buf)
{
	void *retbuf;
	struct s3_get_callback_data data;
	struct s3_get_callback_data_buf *item, *tmp;
	ssize_t offset = 0;

	INIT_LIST_HEAD(&data.bufs); /* make sure the list is initialized */
	data.total_bytes = 0;

	S3BucketContext bucketContext = { bucket, protocolG, uriStyleG,
			accessKeyIdG, secretAccessKeyG };

	S3GetObjectHandler getObjectHandler = { { &responsePropertiesCallback,
			&responseCompleteCallback }, &getObjectDataCallback };

	s3_init();

	do
	{
		S3_get_object(&bucketContext, key, NULL, 0, 0, 0, &getObjectHandler,
				&data);
	} while (S3_status_is_retryable(statusG) && should_retry());

	if (statusG != S3StatusOK)
	{
		printError();
		goto out_err;
	}
	s3_deinit();

	if (data.total_bytes == 0)
	{
		retbuf = NULL;
		offset = 0;
		goto out;
	}
	/* copy the buffers into the return buffer */
	if (!(retbuf = malloc(data.total_bytes)))
	{
		fprintf(stderr, "Malloc error, %s\n", strerror(errno));
		exit(1);
	}

	offset = 0;
	list_for_each_entry_safe(item, tmp, &data.bufs, list)
	{
		memcpy( &(((char *)retbuf)[offset]), item->buf, item->bufsize);
		offset += item->bufsize;
		/* free up the list of buffers */
		list_del(&item->list); /* unchain item */
		free(item->buf);
		free(item);
	}

	out:
	*buf = retbuf; /* return the buffer */
	return offset;
	out_err:
	*buf = NULL;
	return -1;
}

