#include "s3ro.hh"
#include "serialize_to_string.hh"

#define BOOST_ARCHIVE_ITYPE boost::archive::binary_iarchive
#define BOOST_ARCHIVE_OTYPE boost::archive::binary_oarchive

#define PTR_ADD(p,x)                   ((void *)(((unsigned long)(p))+((unsigned long)(x))))

using namespace s3ro;

static string stat_to_str(struct stat *s) {
  return string((char *)s, sizeof(struct stat));
}

template <class T>
string serialize_to_string(T &x) {
  ostringstream ostrm(ios_base::out | ios_base::binary);
  BOOST_ARCHIVE_OTYPE oa(ostrm);
  oa << x;
  return ostrm.str();
}

FileINode::FileINode(HashStore *store, string path) : INode(S3RO_FILE) {
  cout << "put " << path << "\n";
  struct stat attr;
  memset(&attr,0,sizeof(struct stat));

  assert(stat(path.c_str(), &attr) == 0);
  assert(S_ISREG(attr.st_mode));
  
  attr_string = stat_to_str(&attr);

  off_t remaining = attr.st_size;

  if(remaining == 0) 
    return;

  FILE *f = fopen(path.c_str(), "rb");
  assert(f != NULL);
  
  char cbuffer[S3RO_MAX_BLOCK_SIZE];
  void *buffer = (void *) cbuffer;

  while(remaining > 0) {
    ssize_t bytes_read = read(fileno(f), buffer, S3RO_MAX_BLOCK_SIZE);
    remaining -= bytes_read;
    string block(cbuffer, bytes_read);
    string key = store->put(block);
    block_hashes.push_back(key);
    cout << "        put data " << key << "\n";
  }

  fclose(f);
}

DirectoryINode::DirectoryINode(HashStore *store, string path) : INode(S3RO_DIRECTORY) {
  cout << "put " << path << "\n";

  struct stat attr;
  memset(&attr,0,sizeof(struct stat));

  assert(stat(path.c_str(), &attr) == 0);
  assert(S_ISDIR(attr.st_mode));
  
  attr_string = stat_to_str(&attr);
  
  DIR *d = opendir(path.c_str());
  assert(d != NULL);

  struct dirent *ent;

  while( (ent = readdir(d))) {
    if(strcmp(ent->d_name, ".") == 0 || strcmp(ent->d_name, "..") == 0) continue;
    
    string dname = string(ent->d_name);
    string childpath = path + string("/") + dname;

    Key key;
    string hkey;

    if(ent->d_type == DT_DIR) {
      DirectoryINode child_dir(store,childpath);
      hkey = store->put(child_dir.str());
      key = Key(hkey, S3RO_DIRECTORY, ent->d_ino);
      entries[dname] = key;
    } else if(ent->d_type == DT_REG) {
      FileINode child_file(store,childpath);
      hkey = store->put(child_file.str());
      key = Key(hkey, S3RO_FILE, ent->d_ino);
      entries[dname] = key;
    } else if(ent->d_type == DT_LNK) {
      cerr << "symbolic links not currently suppported\n";
      assert(false);
    } else {
      assert(false);  // other file types not implemented
    }
  }
}

string FileINode::str() {
  return serialize_to_string( (const FileINode &) *this);
}

string DirectoryINode::str() {
  return serialize_to_string( (const DirectoryINode &) *this);
}

int S3RO::get_stat(string path, keytype type, struct stat *attr) {
  if(type == S3RO_FILE) {
    FileINode i;
    get<FileINode>(path,i).get_stat(attr);
  } else if(type == S3RO_DIRECTORY) {
    DirectoryINode i;
    get<DirectoryINode>(path,i).get_stat(attr);    
  } else {
    assert(false);
  }
  return 0;
}


void S3RO::read(string path, void *buffer, off_t offset, size_t size) {
  FileINode f;
  get<FileINode>(path,f);
  
  size_t remaining = size;
  off_t maxsize = S3RO_MAX_BLOCK_SIZE;
  vector<string>::iterator i = f.block_hashes.begin();
  
  off_t pos = 0;

  cout << "S3BASE read " << path << "\n";

  while(remaining > 0 && i != f.block_hashes.end()) {
    
    if(pos+maxsize <= offset) {
      i++;
      continue;
    }
    cout << "    fetch data " << *i << "\n";
    string data = store->get(*i);
    off_t soffset = offset - pos;
    size_t to_read = std::min( (size_t) (data.size() - soffset), remaining);
    cout << "         read offset: " << soffset << "  bytes: " << to_read << "\n";
    data.copy((char *)buffer, to_read, soffset);
    offset += to_read;
    pos += to_read;
    remaining -= to_read;
    buffer = PTR_ADD(buffer, to_read);
    i++;
  }
  // if remaining > 0, then more bytes were requested than actually exist
}
