
#include <fstream>
#include <sstream>
#include <string>
#include <iostream>
#include <map>
#include <list>
#include <vector>
#include <set>
#include <utility>
#include <algorithm>

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <pthread.h>

#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/serialization/list.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/map.hpp>
#include <boost/serialization/base_object.hpp>

#define FUSE_USE_VERSION 26

#ifndef _S3RO_INCLUDED
#define _S3RO_INCLUDED

using namespace std;

namespace s3ro {

typedef short keytype;

#define S3RO_MAX_BLOCK_SIZE (1<<16)

const keytype S3RO_FILE = 1;
const keytype S3RO_DIRECTORY = 2;



class HashStore {
public:
  
  // returns data associated with key
  virtual string get(string key) { return string(); }

  // returns hash key
  virtual string put(string data) { return string(); } 

  virtual string put_unless(string data, string key) {
    return string();
  }

  virtual void ls(list<string> &l) {}

};




class INode {
public:
  keytype type;
  INode(keytype type) { this->type = type; }
  virtual string str() { return string();}
};

class Key {
private:
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive &ar, const unsigned int version) 
    {
      ar & key;
      ar & type;
      ar & ino;
    }  

public:
  string key;
  keytype type;
  ino_t ino;

  Key() {};

  Key(string key, keytype type, ino_t ino) {
    this->key = key;
    this->type = type;
    this->ino = ino;
  }
};


class FileINode : public INode {
private:
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive &ar, const unsigned int version) 
    {
      ar & block_hashes;
      ar & attr_string;
    }  


public:
  virtual string str();
  vector<string> block_hashes;
  string attr_string;

  FileINode() : INode(S3RO_FILE) {}
  
  FileINode(HashStore *store, string path);
  // : INode(S3RO_FILE);

  void get_stat(struct stat *attr) { 
    memcpy(attr, attr_string.c_str(), sizeof(struct stat));
  }
  
  void set_stat(struct stat *attr) {
    attr_string = string((char *) attr, sizeof(struct stat));
  }

  void set_inode(ino_t i) {
    struct stat attr;
    get_stat(&attr);
    attr.st_ino = i;
    set_stat(&attr);
  }
};

class DirectoryINode : public INode {
private:
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive &ar, const unsigned int version) 
    {
      ar & entries;
      ar & attr_string;
    }  

public:
  
  map<string,Key> entries;
  string attr_string;

  DirectoryINode() : INode(S3RO_DIRECTORY) {}
  DirectoryINode(HashStore *store, string path);

  virtual string str();

  void get_stat(struct stat *attr) { 
    memcpy(attr, attr_string.c_str(), sizeof(struct stat));
  }

  void set_stat(struct stat *attr) {
    attr_string = string((char *) attr, sizeof(struct stat));
  }
  
  void set_inode(ino_t i) {
    struct stat attr;
    get_stat(&attr);
    attr.st_ino = i;
    set_stat(&attr);
  }

};


class S3RO {
public:
  HashStore *store;
  string root_key;

  S3RO(HashStore *store, string root_key) {
    this->store = store;
    this->root_key = root_key;
  }
  
  S3RO(HashStore *store, string base_path, bool create) {
    // create a new s3ro repository on the given store
    // value of create is immaterial; it's necessary to distinguish it from the other constructor
    this->store = store;
    DirectoryINode root(store, base_path);
    root_key = store->put(root.str());
  }

  int get_stat(string path, keytype type, struct stat *attr);

  static string create(HashStore *store, string path) {
    // returns the store's root directory key
    S3RO s(store, path, true);
    return s.root_key;
  }

  template<class T> 
  T &get(string path, T &obj) { 
    string data = store->get(path);
    istringstream netstrm(data, ios_base::binary | ios_base::in);
    boost::archive::binary_iarchive ia(netstrm);
    ia >> obj;
    return obj;
  }

  void read(string path, void *buffer, off_t offset, size_t size);
  
};

}; // end namespace s3ro




#endif
