
#include "s3ro_base.hh"

extern DFS *dfs;

int S3ROBase::get_stat(string path, struct stat *attr) {
  map<string,s3ro::keytype>::iterator i;
  i = typemap.find(path);

  if(i == typemap.end()) {
    cout <<"path not in typemap: " << path <<"\n";
    assert(false);
  }

  s3ro::keytype t = i->second;
  return base->get_stat(path, t, attr);
}

void S3ROBase::read(void *buffer, string path, off_t offset, size_t size) {
  map<string,s3ro::keytype>::iterator i;
  i = typemap.find(path);
  assert(i != typemap.end());
  s3ro::keytype t = i->second;
  assert(t == s3ro::S3RO_FILE);
  cout << "S3ROBase::read " << path << "\n";
  base->read(path, buffer, offset, size);
}

inode_id S3ROBase::lookup(string directory, string name) {
  map<string,s3ro::keytype>::iterator i;
  i = typemap.find(directory);
  assert(i != typemap.end());
  s3ro::keytype t = i->second;
  assert(t == s3ro::S3RO_DIRECTORY);  
  
  s3ro::DirectoryINode d;
  base->get<s3ro::DirectoryINode>(directory,d);

  map<string,s3ro::Key>::iterator j;
  j = d.entries.find(name);

  if(j == d.entries.end())
    return 0; // not found
  
  s3ro::Key k = j->second;
  record_type(k.key,k.type);
  
  inode_id child = dfs->baptize(k.ino, k.key);
  return child;
}

void S3ROBase::ls(string directory, list<string> &entries) {
  list<ino_t> inos;
  list<string> paths;
  ls(directory, entries, inos, paths);
}

void S3ROBase::ls(string directory, 
		  list<string> &names,
		  list<ino_t> &inos, 
		  list<string> &paths) {

  map<string,s3ro::keytype>::iterator i;
  i = typemap.find(directory);
  assert(i != typemap.end());
  s3ro::keytype t = i->second;
  assert(t == s3ro::S3RO_DIRECTORY);  

  s3ro::DirectoryINode d;
  base->get<s3ro::DirectoryINode>(directory,d);

  map<string,s3ro::Key>::iterator j;
  for(j = d.entries.begin(); j != d.entries.end(); j++) {
    s3ro::Key k = j->second;
    
    record_type(k.key, k.type);

    names.push_back(j->first);
    inos.push_back(k.ino);
    paths.push_back(k.key);
  }
}

void S3ROBase::record_type(string key, s3ro::keytype type) {
  map<string,s3ro::keytype>::iterator i;
  i = typemap.find(key);
  if(i == typemap.end()) {
    (new LogSetS3ROType(key, type))->play();
    //typemap[key] = type;
  }
}

void S3ROBase::set_type(string key, s3ro::keytype type) { 
  typemap[key] = type;
}

void LogUseS3ROBase::replay() {
  BEGIN_REPLAY;
  dfs->base = new S3ROBase(root, bucket, key_id, secret_key);
  END_REPLAY;
}

void LogSetS3ROType::replay() {
  BEGIN_REPLAY;
  dfs->base->set_type(path,type);
  END_REPLAY;
}

ino_t S3ROBase::freeze_get_ino(inode_id i) {
  map<inode_id,ino_t>::iterator j;
  j = reverse_inode_map.find(i);
  if(j != reverse_inode_map.end()) {
    return j->second;
  } else {
    ino_t x = ++max_ino;
    reverse_inode_map[i] = x;
    return x;
  }
}

Event *S3ROBase::freeze() {
  cout << "Freezing...\n";

  // crawl complete filesystem to map all inodes
  // (not efficient!)
  cout << "Crawling...\n";
  dfs->crawl();

  cout << "Commit pending log...\n";
  dfs->commit_pending_log(true);

  
  reverse_inode_map.clear();
  max_ino = 0;
  
  cout << "Build reverse inode map\n";

  map<ino_t,inode_id>::iterator ii;
  for(ii = dfs->inode_map.begin(); ii != dfs->inode_map.end(); ii++) {
    reverse_inode_map[ii->second] = ii->first;
    max_ino = std::max(max_ino, ii->first);
  }
  
  //pthread_mutex_lock(&dfs->pending_lock);

  cout << "Freezing summaries...\n";

  string new_root;

  new_root = freeze_dir_summary(dfs->get_root_summary());
  cout << "freeze: new root " << new_root << "\n";
  dfs->commit_pending_log(true);

  Event *newbase = new LogUseS3ROBase(keyid, secret_key, bucket, new_root);

  dfs->nuke();
  cout << "Freeze done!\n";
  return newbase;
  //pthread_mutex_unlock(&dfs->pending_lock);
}

string S3ROBase::freeze_summary(Summary *s) {
  if(s->type == FILE_SUMMARY) {
    return freeze_file_summary( (FileSummary *) s);
  } else if(s->type == DIRECTORY_SUMMARY) {
    return freeze_dir_summary( (DirectorySummary *) s);
  } else if(s->type == STUB_SUMMARY) {
    return s->base_path; // no change
  } else {
    assert(false);
  }
}

string S3ROBase::freeze_file_summary(FileSummary *s) {
  cout << "   freeze file " << s->my_id << " ("<< s->base_path<<")\n";
  
  s3ro::FileINode f;

  // inefficient, should really look at overlays
  if(s->in_base)
    base->get<s3ro::FileINode>(s->base_path, f);    

  if(s->mstat != NULL)
    f.set_stat(s->mstat);

  f.set_inode(freeze_get_ino(s->my_id));

  struct stat attr;
  f.get_stat(&attr);

  if(s->content_diff.size() > 0) {
    // content may have changed,
    // re-create block hashes
    f.block_hashes.clear();
    
    off_t remaining = attr.st_size;
    
    char cbuffer[S3RO_MAX_BLOCK_SIZE];
    void *buffer = (void *) cbuffer;

    off_t offset = 0;

    while(remaining > 0) {
      ssize_t to_read = std::min( (off_t) S3RO_MAX_BLOCK_SIZE, remaining);
      ssize_t bytes_read = s->read( (char *) buffer, offset, to_read);

      assert(to_read == bytes_read);

      remaining -= bytes_read;
      offset += bytes_read;
      string block(cbuffer, bytes_read);
      string key = base->store->put(block);
      f.block_hashes.push_back(key);
      cout << "        put data " << key << "\n";
    }
    
  }
  
  return (s->in_base ?
	  base->store->put_unless(f.str(), s->base_path) :
	  base->store->put(f.str()));
}

string S3ROBase::freeze_dir_summary(DirectorySummary *s) {
  cout << "   freeze dir " << s->my_id << " ("<< s->base_path<<")\n";
  s3ro::DirectoryINode d;

  if(s->in_base)
    base->get<s3ro::DirectoryINode>(s->base_path, d);


  if(s->mstat != NULL)
    d.set_stat(s->mstat);
  
  d.set_inode(freeze_get_ino(s->my_id));
    
  // update entries
  map<string,inode_id>::iterator ddi;
  for(ddi = s->dirent_diff.begin(); ddi != s->dirent_diff.end(); ddi++) {
    
    string name = ddi->first;
    inode_id child = ddi->second;
    
    map<string,s3ro::Key>::iterator e;
    e = d.entries.find(name);
    if(e != d.entries.end()) 
      d.entries.erase(e);
   
    if(child == 0 || child == DELETED)
      continue;

    Summary *csum = dfs->get_summary(child, false);
    assert(csum != NULL);

    s3ro::keytype t;
    if(csum->type == DIRECTORY_SUMMARY)
      t = s3ro::S3RO_DIRECTORY;
    else if(csum->type == FILE_SUMMARY) 
      t = s3ro::S3RO_FILE;
    else 
      assert(false);

    string key = freeze_summary(csum); // freeze the child
    
    d.entries[name] = s3ro::Key(key, t, freeze_get_ino(child));
  }
  
  return (s->in_base ?
	  base->store->put_unless(d.str(), s->base_path) :
	  base->store->put(d.str()));
  
}
